#include <stdio.h>
#include <strings.h>
#include <iostream.h>
#include <fstream.h>
#include "suffix_tree.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ddi.h> // max
#include <sys/time.h>
#include <assert.h>

extern ofstream StatisticsFileStream;

suffix_tree_node* new_leaf_node() {
  suffix_tree_node* node = new suffix_tree_node;
  node->doc_count = 0;
  node->is_leaf = true;
  node->path = NULL;
  node->u.leaf.num_paths = 0;
  node->u.leaf.super_paths = NULL;
  return node;
}

suffix_tree_node* new_non_leaf_node() {
  suffix_tree_node* node = new suffix_tree_node;
  node->doc_count = 0;
  node->is_leaf = false;
  node->path = NULL;
  node->u.non_leaf.num_children = 0;
  node->u.non_leaf.labels = NULL;
  node->u.non_leaf.children = NULL;
  return node;
}

// reverse_path returns the reverse of a path
// eg: if path_ is "a.bd.c" then the fn returns c.bd.a
//     if path_ is "" then the fn returns ""
//     if path_ is "a.cd.x." the fn returns wrong value (error case)
//     ('.' cannot terminate/begin a path)
char *reverse_path(char *path_) {
  int uninit_len = strlen(path_), token_len;
  char *path_copy = strdup(path_);
  char *reverse = new char[uninit_len + 1];
  char *token;
  bool dot_flag = false;

  strcpy(reverse, path_copy);
  token = strtok(path_copy, ".");
  while(token) {
    if(dot_flag) {reverse[uninit_len-1] = '.'; uninit_len--;}
    else dot_flag = true;
      
    token_len = strlen(token);
    strncpy(reverse + uninit_len - token_len, token, token_len);
    uninit_len -= token_len;
    token = strtok(NULL, ".");
  }
  free(path_copy);
  return reverse;
}

bool my_compare(char *node_string_, char *new_string_, 
		char *&common_, char *&node_left_ , char *&new_left_) {  
  int node_len = strlen(node_string_);
  int new_len = strlen(new_string_);
  int min_len = (node_len < new_len ? node_len : new_len);
  int max_len = (node_len > new_len ? node_len : new_len);
  char *token1, *token2;
  int i1, i2, start1, start2;
  bool break_flag;

  common_ = new char[min_len + 1];
  strcpy(common_, "");

  break_flag = false;

  i1 = 0;
  start1 = i1;
  while(node_string_[i1] != '\0' && node_string_[i1] != '.') i1++;
  token1 = new char[max_len + 1];
  strncpy(token1, node_string_+start1, i1-start1);
  token1[i1-start1] = 0;

  i2 = 0;
  start2 = i2;
  while(new_string_[i2] != '\0' && new_string_[i2] != '.') i2++;
  token2 = new char[max_len + 1];
  strncpy(token2, new_string_+start2, i2-start2);
  token2[i2-start2] = 0;

  while(strcmp(token1, token2) == 0) {
    if(strcmp(common_, "") != 0) strcat(common_, ".");
    strcat(common_, token1);

    if(node_string_[i1] == '\0' || new_string_[i2] == '\0') break_flag = true;
    if(node_string_[i1] == '\0') start1 = i1;
    if(new_string_[i2] == '\0') start2 = i2;
    if(node_string_[i1] == '.') { i1++; start1 = i1; }
    if(new_string_[i2] == '.') { i2++; start2 = i2; }
    if(break_flag) break;

    while(node_string_[i1] != '\0' && node_string_[i1] != '.') i1++;
    strncpy(token1, node_string_+start1, i1-start1);
    token1[i1-start1] = 0;

    while(new_string_[i2] != '\0' && new_string_[i2] != '.') i2++;
    strncpy(token2, new_string_+start2, i2-start2);    
    token2[i2-start2] = 0;
  }
  node_left_ = new char[node_len-start1+1];
  strncpy(node_left_, node_string_ + start1, node_len-start1);
  node_left_[node_len-start1] = 0;

  new_left_ = new char[new_len-start2+1];
  strncpy(new_left_, new_string_ + start2, new_len-start2);
  new_left_[new_len-start2] = 0;

  delete []token1;
  delete []token2;
  if(strlen(common_) == 0) return false;
  return true;
}

void insert_path(suffix_tree_node *&root_, char *path_, char *sub_path_, char *super_path_, 
		 int file_off_, int num_docs_) {
  bool found;
  char *node_left, *new_path, *new_left, *common;
  suffix_tree_node *inter_node, *temp_node;
  char **temp_labels, **temp_super_paths;
  int *temp_file_offs, *temp_num_docs, path_len, inter_node_len;
  suffix_tree_node **temp_children;
  int i, mark, max_num_docs;
  
  if(root_ == NULL) {
    root_ = new_non_leaf_node();

    path_len = strlen(path_);
    root_->path = new char[path_len + 1];
    strcpy(root_->path, path_);
    
    root_->u.non_leaf.num_children = 1;

    root_->u.non_leaf.labels = new (char *)[1];
    root_->u.non_leaf.labels[0] = new char[strlen(sub_path_) + 1];
    strcpy(root_->u.non_leaf.labels[0], sub_path_);

    root_->u.non_leaf.children = new (suffix_tree_node *)[1];  
    temp_node = new_leaf_node();
    temp_node->path = new char[path_len + 1 + strlen(sub_path_) + 1];
    if(path_len > 0)
      sprintf(temp_node->path, "%s.%s", path_, sub_path_);
    else
      strcpy(temp_node->path, sub_path_);
    temp_node->u.leaf.num_paths = 1;

    temp_node->u.leaf.super_paths = new (char *)[temp_node->u.leaf.num_paths];
    temp_node->u.leaf.super_paths[0] = new (char)[strlen(super_path_) + 1];
    strcpy(temp_node->u.leaf.super_paths[0], super_path_);

    temp_node->u.leaf.file_offs = new (int)[temp_node->u.leaf.num_paths];
    temp_node->u.leaf.file_offs[0] = file_off_;

    temp_node->u.leaf.num_docs = new (int)[temp_node->u.leaf.num_paths];
    temp_node->u.leaf.num_docs[0] = num_docs_;

    root_->u.non_leaf.children[0] = temp_node;
    
    root_->u.non_leaf.num_docs = new (int)[1];
    root_->u.non_leaf.num_docs[0] = num_docs_;
    return;
  }
    
  if(root_->is_leaf) {
    if(strcmp(sub_path_, "") == 0) {
      for(i=0; 
	  ((i<root_->u.leaf.num_paths) && (strcmp(root_->u.leaf.super_paths[i], super_path_) < 0)); 
	  i++);
      mark = i;
      temp_super_paths = root_->u.leaf.super_paths;
      temp_file_offs = root_->u.leaf.file_offs;
      temp_num_docs = root_->u.leaf.num_docs;
      root_->u.leaf.num_paths++;

      root_->u.leaf.super_paths = new (char *)[root_->u.leaf.num_paths];
      root_->u.leaf.file_offs = new (int)[root_->u.leaf.num_paths];
      root_->u.leaf.num_docs = new (int)[root_->u.leaf.num_paths];

      for(i=0; i<mark; i++) {
	root_->u.leaf.super_paths[i] = temp_super_paths[i];
	root_->u.leaf.file_offs[i] = temp_file_offs[i];
	root_->u.leaf.num_docs[i] = temp_num_docs[i];
      }
      root_->u.leaf.super_paths[i] = new char[strlen(super_path_) + 1];
      strcpy(root_->u.leaf.super_paths[i], super_path_);
      root_->u.leaf.file_offs[i] = file_off_;
      root_->u.leaf.num_docs[i]  = num_docs_;
      i++;
      for(int j=i; j<(root_->u.leaf.num_paths); j++) {
	root_->u.leaf.super_paths[j] = temp_super_paths[j-1];
	root_->u.leaf.file_offs[j] = temp_file_offs[j-1];
	root_->u.leaf.num_docs[j] = temp_num_docs[j-1];
      }
      delete []temp_super_paths;
      delete []temp_file_offs;
      delete []temp_num_docs;
    }
    else {
      temp_node = new_leaf_node();
      path_len = strlen(root_->path);
      temp_node->path = new char[path_len + 1];
      strcpy(temp_node->path, root_->path);
      temp_node->u.leaf.num_paths = root_->u.leaf.num_paths;
      temp_node->u.leaf.super_paths = new (char *)[temp_node->u.leaf.num_paths];
      temp_node->u.leaf.file_offs = new (int)[temp_node->u.leaf.num_paths];
      temp_node->u.leaf.num_docs = new (int)[temp_node->u.leaf.num_paths];
      max_num_docs = root_->u.leaf.num_docs[0];
      for(i=0; i<(temp_node->u.leaf.num_paths); i++) {
	temp_node->u.leaf.super_paths[i] = new char[strlen(root_->u.leaf.super_paths[i]) + 1];
	strcpy(temp_node->u.leaf.super_paths[i], root_->u.leaf.super_paths[i]);
	temp_node->u.leaf.file_offs[i] = root_->u.leaf.file_offs[i];
	temp_node->u.leaf.num_docs[i] = root_->u.leaf.num_docs[i];
	if(temp_node->u.leaf.num_docs[i] > max_num_docs) max_num_docs = temp_node->u.leaf.num_docs[i];
	delete []root_->u.leaf.super_paths[i];
      }
      delete []root_->u.leaf.super_paths;
      delete []root_->u.leaf.file_offs;
      delete []root_->u.leaf.num_docs;
      root_->is_leaf = false;
      root_->u.non_leaf.num_children = 2;
      root_->u.non_leaf.labels = new (char *)[root_->u.non_leaf.num_children];
      root_->u.non_leaf.labels[0] = new char[1];
      strcpy(root_->u.non_leaf.labels[0], "");
      root_->u.non_leaf.labels[1] = new char[strlen(sub_path_) + 1];
      strcpy(root_->u.non_leaf.labels[1], sub_path_);
      root_->u.non_leaf.children = new (suffix_tree_node *)[root_->u.non_leaf.num_children];
      root_->u.non_leaf.children[0] = temp_node;
      
      temp_node = new_leaf_node();
      temp_node->path = new char[path_len + 1 + strlen(sub_path_) + 1];      
      if(path_len > 0)
	sprintf(temp_node->path, "%s.%s", root_->path, sub_path_);
      else
	strcpy(temp_node->path, sub_path_);

      temp_node->u.leaf.num_paths = 1;
      temp_node->u.leaf.super_paths = new (char *)[temp_node->u.leaf.num_paths];
      temp_node->u.leaf.file_offs = new (int)[temp_node->u.leaf.num_paths];
      temp_node->u.leaf.num_docs = new (int)[temp_node->u.leaf.num_paths];
      temp_node->u.leaf.super_paths[0] = new (char)[strlen(super_path_) + 1];
      strcpy(temp_node->u.leaf.super_paths[0], super_path_);
      temp_node->u.leaf.file_offs[0] = file_off_;
      temp_node->u.leaf.num_docs[0] = num_docs_;
      root_->u.non_leaf.children[1] = temp_node;
      
      root_->u.non_leaf.num_docs = new (int)[root_->u.non_leaf.num_children];
      root_->u.non_leaf.num_docs[0] = max_num_docs;
      root_->u.non_leaf.num_docs[1] = num_docs_;
    }
    return;
  }

  if(strcmp(sub_path_, "") == 0) {
    if(root_->u.non_leaf.num_children >= 1) {
      if(strcmp(root_->u.non_leaf.labels[0], "") == 0) {
	if(num_docs_ > root_->u.non_leaf.num_docs[0]) root_->u.non_leaf.num_docs[0] = num_docs_;
	return insert_path(root_->u.non_leaf.children[0], path_, sub_path_, super_path_, file_off_, num_docs_);
      }
    }
    temp_labels = root_->u.non_leaf.labels;
    root_->u.non_leaf.num_children++;
    root_->u.non_leaf.labels = new (char *)[root_->u.non_leaf.num_children];
    root_->u.non_leaf.labels[0] = new (char)[1];
    strcpy(root_->u.non_leaf.labels[0], "");
    for(i=1; i<root_->u.non_leaf.num_children; i++)
      root_->u.non_leaf.labels[i] = temp_labels[i-1];
    delete []temp_labels;
    
    temp_children = root_->u.non_leaf.children;
    root_->u.non_leaf.children = new (suffix_tree_node *)[root_->u.non_leaf.num_children];
    temp_node = new_leaf_node();
    temp_node->path = new char[strlen(path_) + 1];
    strcpy(temp_node->path, path_);
    temp_node->u.leaf.num_paths = 1;
    temp_node->u.leaf.super_paths = new (char *)[temp_node->u.leaf.num_paths];
    temp_node->u.leaf.file_offs = new (int)[temp_node->u.leaf.num_paths];
    temp_node->u.leaf.num_docs = new (int)[temp_node->u.leaf.num_paths];

    temp_node->u.leaf.super_paths[0] = new (char)[strlen(super_path_) + 1];
    strcpy(temp_node->u.leaf.super_paths[0], super_path_);
    temp_node->u.leaf.file_offs[0] = file_off_;
    temp_node->u.leaf.num_docs[0] = num_docs_;

    root_->u.non_leaf.children[0] = temp_node;
    for(i=1; i<root_->u.non_leaf.num_children; i++)
      root_->u.non_leaf.children[i] = temp_children[i-1];
    delete []temp_children;

    temp_num_docs = root_->u.non_leaf.num_docs;
    root_->u.non_leaf.num_docs = new (int)[root_->u.non_leaf.num_children];
    root_->u.non_leaf.num_docs[0] = num_docs_;
    for(i=1; i<root_->u.non_leaf.num_children; i++)
      root_->u.non_leaf.num_docs[i] = temp_num_docs[i-1];
    delete []temp_num_docs;
    return;
  }

  path_len = strlen(path_);
  found = false;
  for(i = 0; !found && (i < root_->u.non_leaf.num_children); i++) {
    if(strcmp(root_->u.non_leaf.labels[i], "") == 0) continue;
    if(my_compare(root_->u.non_leaf.labels[i], sub_path_, common, node_left, new_left)) {
      if(strcmp(root_->u.non_leaf.labels[i], common) == 0) {
	delete []common;
	delete []node_left;
	new_path = new char[path_len + 1 + strlen(root_->u.non_leaf.labels[i]) + 1];
	if(path_len > 0)
	  sprintf(new_path, "%s.%s", path_, root_->u.non_leaf.labels[i]);	
	else
	  strcpy(new_path, root_->u.non_leaf.labels[i]);	
	insert_path(root_->u.non_leaf.children[i], new_path, new_left, super_path_, file_off_, num_docs_);
	if(num_docs_ > root_->u.non_leaf.num_docs[i]) root_->u.non_leaf.num_docs[i] = num_docs_;
	delete []new_path;
	delete []new_left;
      }
      else {
	inter_node = new_non_leaf_node();
	inter_node->path = new char[path_len + 1 + strlen(common) + 1];
	if(path_len > 0)
	  sprintf(inter_node->path, "%s.%s", path_, common);
	else
	  strcpy(inter_node->path, common);
	inter_node->u.non_leaf.num_children = 2;
	inter_node->u.non_leaf.labels = new (char *)[2];
	inter_node->u.non_leaf.children = new (suffix_tree_node *)[2];
	inter_node->u.non_leaf.num_docs = new (int) [2];

	temp_node = new_leaf_node();
	inter_node_len = strlen(inter_node->path);
	temp_node->path = new char[inter_node_len + 1 + strlen(new_left) + 1];
	if(inter_node_len > 0)
	  sprintf(temp_node->path, "%s.%s", inter_node->path, new_left);
	else
	  strcpy(temp_node->path, new_left);
	temp_node->u.leaf.num_paths = 1;
	temp_node->u.leaf.super_paths = new (char *)[temp_node->u.leaf.num_paths];
	temp_node->u.leaf.file_offs = new (int)[temp_node->u.leaf.num_paths];
	temp_node->u.leaf.num_docs = new (int)[temp_node->u.leaf.num_paths];
	
	temp_node->u.leaf.super_paths[0] = new (char)[strlen(super_path_) + 1];
	strcpy(temp_node->u.leaf.super_paths[0], super_path_);
	temp_node->u.leaf.file_offs[0] = file_off_;
	temp_node->u.leaf.num_docs[0] = num_docs_;

	if(strcmp(new_left, node_left) < 0){
	  inter_node->u.non_leaf.labels[0] = new_left;
	  inter_node->u.non_leaf.labels[1] = node_left;
	  inter_node->u.non_leaf.children[0] = temp_node;
	  inter_node->u.non_leaf.children[1] = root_->u.non_leaf.children[i];  
	  inter_node->u.non_leaf.num_docs[0] = num_docs_;
	  inter_node->u.non_leaf.num_docs[1] = root_->u.non_leaf.num_docs[i];
	}
	else {
	  inter_node->u.non_leaf.labels[0] = node_left;
	  inter_node->u.non_leaf.labels[1] = new_left;
	  inter_node->u.non_leaf.children[0] = root_->u.non_leaf.children[i];
	  inter_node->u.non_leaf.children[1] = temp_node;
	  inter_node->u.non_leaf.num_docs[0] = root_->u.non_leaf.num_docs[i];
	  inter_node->u.non_leaf.num_docs[1] = num_docs_;
	}
	delete []root_->u.non_leaf.labels[i];
	root_->u.non_leaf.labels[i] = common;
	root_->u.non_leaf.children[i] = inter_node;
	if(num_docs_ > root_->u.non_leaf.num_docs[i]) root_->u.non_leaf.num_docs[i] = num_docs_;
      }
      found = true;
    }
    else {
      delete []common;
      delete []node_left;
      delete []new_left;
    }
  }
  
  if(!found) {
    for(i=0; i<(root_->u.non_leaf.num_children) && (strcmp(root_->u.non_leaf.labels[i], sub_path_) < 0); i++);
    mark = i;
    temp_labels = root_->u.non_leaf.labels;
    temp_children = root_->u.non_leaf.children;
    temp_num_docs = root_->u.non_leaf.num_docs;

    (root_->u.non_leaf.num_children)++;
    root_->u.non_leaf.labels = new (char *)[root_->u.non_leaf.num_children];
    root_->u.non_leaf.children = new (suffix_tree_node *)[root_->u.non_leaf.num_children];
    root_->u.non_leaf.num_docs = new (int)[root_->u.non_leaf.num_children];
    for(i=0; i<mark; i++) {
      root_->u.non_leaf.labels[i] = temp_labels[i];
      root_->u.non_leaf.children[i] = temp_children[i];
      root_->u.non_leaf.num_docs[i] = temp_num_docs[i];
    }
    root_->u.non_leaf.labels[mark] = new char[strlen(sub_path_) + 1];
    strcpy(root_->u.non_leaf.labels[mark], sub_path_);
    temp_node = new_leaf_node();
    path_len = strlen(root_->path);
    temp_node->path = new char[path_len + 1 + strlen(sub_path_) + 1];
    if(path_len > 0)
      sprintf(temp_node->path, "%s.%s", root_->path, sub_path_);
    else
      strcpy(temp_node->path, sub_path_);
    temp_node->u.leaf.num_paths = 1;
    temp_node->u.leaf.super_paths = new (char *)[temp_node->u.leaf.num_paths];
    temp_node->u.leaf.file_offs = new (int)[temp_node->u.leaf.num_paths];
    temp_node->u.leaf.num_docs = new (int)[temp_node->u.leaf.num_paths];

    temp_node->u.leaf.super_paths[0] = new (char)[strlen(super_path_) + 1];
    strcpy(temp_node->u.leaf.super_paths[0], super_path_);
    temp_node->u.leaf.file_offs[0] = file_off_;
    temp_node->u.leaf.num_docs[0] = num_docs_;
    
    root_->u.non_leaf.children[mark] = temp_node;
    root_->u.non_leaf.num_docs[mark] = num_docs_;

    for(i=(mark+1); i<(root_->u.non_leaf.num_children); i++) {
      root_->u.non_leaf.labels[i] = temp_labels[i-1];
      root_->u.non_leaf.children[i] = temp_children[i-1];
      root_->u.non_leaf.num_docs[i] = temp_num_docs[i-1];
    }
    delete []temp_labels;
    delete []temp_children;
    delete []temp_num_docs;
  }  
  return;
}

void delete_suffix_tree(suffix_tree_node *&root_) {
  if(root_) {
    delete []root_->path;
    if(root_->is_leaf == false) {
      for(int i=0; i<root_->u.non_leaf.num_children; i++) {
	delete []root_->u.non_leaf.labels[i];
	delete_suffix_tree(root_->u.non_leaf.children[i]);
      }
      delete []root_->u.non_leaf.labels;
      delete []root_->u.non_leaf.children;
      delete []root_->u.non_leaf.num_docs;
      delete root_;
    }
    else {
      for(int i=0; i<root_->u.leaf.num_paths; i++) {
	delete []root_->u.leaf.super_paths[i];
      }
      delete []root_->u.leaf.super_paths;
      delete []root_->u.leaf.file_offs;
      delete []root_->u.leaf.num_docs;
      delete root_;
    }
    root_ = NULL;
  }
}

void print_to_file_help(int st_file_id_, int st_index_file_id_, int &offset_, suffix_tree_node *root_) {
  int is_leaf, super_path_len, label_len, path_len, offset;
  if(root_ == NULL) return;
  is_leaf = (root_->is_leaf ? 1 : 0);

  // write node to st file  
  offset = offset_;
  path_len = strlen(root_->path);
  if(write(st_file_id_, (char *)&path_len, sizeof(int)) < sizeof(int)) return;
  if(write(st_file_id_, (char *)(root_->path), path_len) < path_len) return;
  if(write(st_file_id_, (char *)&is_leaf, sizeof(int)) < sizeof(int)) return;
  offset_ += (sizeof(int) + path_len + sizeof(int));

  if(is_leaf == 1) {
    if(write(st_file_id_, (char *)&(root_->u.leaf.num_paths), sizeof(int)) < sizeof(int)) return;
    offset_ += sizeof(int);
    for(int i=0; i<(root_->u.leaf.num_paths); i++) {
      super_path_len = strlen(root_->u.leaf.super_paths[i]);
      if(write(st_file_id_, (char *)&super_path_len, sizeof(int)) < sizeof(int)) return;
      if(write(st_file_id_, (char *)(root_->u.leaf.super_paths[i]), super_path_len) < super_path_len) return;
      if(write(st_file_id_, (char *)&(root_->u.leaf.file_offs[i]), sizeof(int)) < sizeof(int)) return;
      if(write(st_file_id_, (char *)&(root_->u.leaf.num_docs[i]), sizeof(int)) < sizeof(int)) return;
      offset_ += (3 * sizeof(int) + super_path_len);
    }
  }
  else {
    if(write(st_file_id_, (char *)&(root_->u.non_leaf.num_children), sizeof(int)) < sizeof(int)) return;
    offset_ += sizeof(int);

    for(int i=0; i<(root_->u.non_leaf.num_children); i++) {
      label_len = strlen(root_->u.non_leaf.labels[i]);

      // write offsets to index file  
      if(write(st_index_file_id_, (char *)&offset, sizeof(int)) < sizeof(int)) return;
      if(write(st_index_file_id_, (char *)&offset_, sizeof(int)) < sizeof(int)) return;

      if(write(st_file_id_, (char *)&label_len, sizeof(int)) < sizeof(int)) return;
      if(write(st_file_id_, (char *)(root_->u.non_leaf.labels[i]), label_len) < label_len) return;
      offset_ += (sizeof(int) + label_len);
      
      print_to_file_help(st_file_id_, st_index_file_id_, offset_, root_->u.non_leaf.children[i]);      
    }
  }
}

void print_to_file(char *st_file_, suffix_tree_node *root_) {
  char *st_index_file = new char[strlen(st_file_) + 1 + strlen(".idx")];
  sprintf(st_index_file, "%s.idx", st_file_);
  int st_file_id = open(st_file_, O_CREAT | O_WRONLY | O_TRUNC, S_IREAD | S_IWRITE);
  int st_index_file_id = open(st_index_file, O_CREAT | O_WRONLY | O_TRUNC, S_IREAD | S_IWRITE);
  int offset = 0;
  print_to_file_help(st_file_id, st_index_file_id, offset, root_);
  close(st_file_id);
  close(st_index_file_id);
  delete []st_index_file;
}


void read_from_file_help(int st_file_id_, suffix_tree_node *&root_) {
  int path_len, is_leaf, label_len, super_path_len;
  char *path;

  if(read(st_file_id_, (char *)&path_len, sizeof(int)) < sizeof(int)) return;
  path = new char[path_len + 1];
  if(read(st_file_id_, (char *) path, path_len) < path_len) return;
  path[path_len] = 0;
  
  if(read(st_file_id_, (char *)&is_leaf, sizeof(int)) < sizeof(int)) return;
  if(is_leaf) {
    if(root_ == NULL) root_ = new_leaf_node();   
    root_->path = path;
    root_->is_leaf = true;
    if(read(st_file_id_, (char *)&(root_->u.leaf.num_paths), sizeof(int)) < sizeof(int)) return;
    root_->u.leaf.super_paths = new (char *)[root_->u.leaf.num_paths];
    root_->u.leaf.file_offs = new (int)[root_->u.leaf.num_paths];
    root_->u.leaf.num_docs = new (int)[root_->u.leaf.num_paths];
    for(int i=0; i<(root_->u.leaf.num_paths); i++) {
      if(read(st_file_id_, (char *)&super_path_len, sizeof(int)) < sizeof(int)) return;
      root_->u.leaf.super_paths[i] = new char[super_path_len + 1];
      if(read(st_file_id_, (char *)root_->u.leaf.super_paths[i], super_path_len) < super_path_len) return;
      root_->u.leaf.super_paths[i][super_path_len] = 0;
      if(read(st_file_id_, (char *)&(root_->u.leaf.file_offs[i]), sizeof(int)) < sizeof(int)) return;
      if(read(st_file_id_, (char *)&(root_->u.leaf.num_docs[i]), sizeof(int)) < sizeof(int)) return;
    }
  }
  else {
    if(root_ == NULL) root_ = new_non_leaf_node();   
    root_->path = path;
    root_->is_leaf = false;
    if(read(st_file_id_, (char *)&(root_->u.non_leaf.num_children), sizeof(int)) < sizeof(int)) return;
    root_->u.non_leaf.labels = new (char *)[root_->u.non_leaf.num_children];
    root_->u.non_leaf.children = new (suffix_tree_node *)[root_->u.non_leaf.num_children];
    for(int i=0; i<(root_->u.non_leaf.num_children); i++) {
      if(read(st_file_id_, (char *)&label_len, sizeof(int)) < sizeof(int)) return;
      root_->u.non_leaf.labels[i] = new char[label_len + 1];
      if(read(st_file_id_, (char *)root_->u.non_leaf.labels[i], label_len) < label_len) return;
      root_->u.non_leaf.labels[i][label_len] = 0;
      root_->u.non_leaf.children[i] = 0;
      read_from_file_help(st_file_id_, root_->u.non_leaf.children[i]);
    }
  }
}

// creates a suffix tree (root_) from data stored in file (st_file_)
void read_from_file(char *st_file_, suffix_tree_node *&root_) {
  int st_file_id;

  st_file_id = open(st_file_, O_RDONLY);
  read_from_file_help(st_file_id, root_);

  close(st_file_id);
}

void zoom_help(int st_file_id_, int path_file_id_, char *root_path_, char *root_label_, 
	       suffix_tree_node *&root_same_, suffix_tree_node *&root_other_, int &visited_count_) {
  int root_path_len, bpath_len, is_leaf, label_len, super_path_len, path_len;
  int num_paths, file_off, num_docs, num_children, root_label_len;
  char *bpath, *super_path, *label, *new_super_path, *same_sub, *other_sub, *other_super, *temp, *path_ptr;

  if(read(st_file_id_, (char *)&bpath_len, sizeof(int)) < sizeof(int)) return;
  bpath = new char[bpath_len + 1];
  if(read(st_file_id_, (char *) bpath, bpath_len) < bpath_len) return;
  bpath[bpath_len] = 0;
  
  root_path_len = strlen(root_path_);
  root_label_len = strlen(root_label_);

  if(read(st_file_id_, (char *)&is_leaf, sizeof(int)) < sizeof(int)) return;
  visited_count_++;
  if(is_leaf) {
    if(read(st_file_id_, (char *)&num_paths, sizeof(int)) < sizeof(int)) return;
    for(int i=0; i<num_paths; i++) {
      if(read(st_file_id_, (char *)&super_path_len, sizeof(int)) < sizeof(int)) return;
      super_path = new char[super_path_len + 1];
      if(read(st_file_id_, (char *)super_path, super_path_len) < super_path_len) return;
      super_path[super_path_len] = 0;
      if(read(st_file_id_, (char *)&file_off, sizeof(int)) < sizeof(int)) return;
      if(read(st_file_id_, (char *)&num_docs, sizeof(int)) < sizeof(int)) return;

      if(super_path_len > 0) {
	path_len = super_path_len + 1 + bpath_len;
	path_ptr = new char[path_len + 1];
	sprintf(path_ptr, "%s.%s", super_path, bpath);	
	if(write(path_file_id_, (char *)&path_len, sizeof(int)) < sizeof(int)) return;
	if(write(path_file_id_, (char *) path_ptr, path_len) < path_len) return;
	delete []path_ptr;
      }
      else { 
	if(write(path_file_id_, (char *)&bpath_len, sizeof(int)) < sizeof(int)) return;
	if(write(path_file_id_, (char *) bpath, bpath_len) < bpath_len) return;
      }
      if(write(path_file_id_, (char *)&file_off, sizeof(int)) < sizeof(int)) return;
      if(write(path_file_id_, (char *)&num_docs, sizeof(int)) < sizeof(int)) return;

      if(super_path_len > 0 ) {
	if(root_path_len > 0) {
	  same_sub = bpath + root_path_len;
	  if(same_sub[0] == '.') same_sub++;

	  new_super_path = new char[super_path_len + 1 + root_path_len + 1];
	  sprintf(new_super_path, "%s.%s", super_path, root_path_);
	  insert_path(root_same_, "", same_sub, new_super_path, file_off, num_docs);

	  other_sub = reverse_path(new_super_path);
	  if(root_label_len > 0) {
	    temp = new char[strlen(other_sub) + 1 + root_label_len + 1];
	    char *reverse_root_label = reverse_path(root_label_);
	    sprintf(temp, "%s.%s", reverse_root_label, other_sub);
	    delete []reverse_root_label;
	    delete []other_sub;
	    other_sub = temp;
	  }

	  char *other_super_ptr = same_sub + root_label_len;
	  if(other_super_ptr[0] == '.') other_super_ptr++;
	  other_super = reverse_path(other_super_ptr);

	  insert_path(root_other_, "", other_sub, other_super, file_off, num_docs);

	  delete []new_super_path;
	}
	else {
	  same_sub = bpath;
	  insert_path(root_same_, "", same_sub, super_path, file_off, num_docs);

	  other_sub = reverse_path(super_path);
	  if(root_label_len > 0) {
	    temp = new char[strlen(other_sub) + 1 + root_label_len + 1];
	    char *reverse_root_label = reverse_path(root_label_);
	    sprintf(temp, "%s.%s", reverse_root_label, other_sub);
	    delete []reverse_root_label;
	    delete []other_sub;
	    other_sub = temp;
	  }

	  char *other_super_ptr = same_sub + root_label_len;
	  if(other_super_ptr[0] == '.') other_super_ptr++;
	  other_super = reverse_path(other_super_ptr);

	  insert_path(root_other_, "", other_sub, other_super, file_off, num_docs);
	}
      }
      else {
	same_sub = bpath + root_path_len;
	if(same_sub[0] == '.') same_sub++;
	insert_path(root_same_, "", same_sub, root_path_, file_off, num_docs);
	
	if(root_label_len > 0) {
	  if(root_path_len > 0) {
	    temp = reverse_path(root_path_);
	    other_sub = new char[root_path_len + 1 + root_label_len + 1];	    
	    char *reverse_root_label = reverse_path(root_label_);
	    sprintf(other_sub, "%s.%s", reverse_root_label, temp);
	    delete []reverse_root_label;
	    delete []temp;
	  }
	  else {
	    other_sub = reverse_path(root_label_);
	  }
	}
	else {
	  other_sub = reverse_path(root_path_);
	}

	char *other_super_ptr = same_sub + root_label_len;
	if(other_super_ptr[0] == '.') other_super_ptr++;
	other_super = reverse_path(other_super_ptr);
	
	insert_path(root_other_, "", other_sub, other_super, file_off, num_docs);
      }

      
      delete []super_path;
      delete []other_sub;
      delete []other_super;
    }
  }
  else {
    if(read(st_file_id_, (char *)&num_children, sizeof(int)) < sizeof(int)) return;
    for(int i=0; i<num_children; i++) {
      if(read(st_file_id_, (char *)&label_len, sizeof(int)) < sizeof(int)) return;
      label = new char[label_len + 1];
      if(read(st_file_id_, (char *)label, label_len) < label_len) return;
      label[label_len] = 0;
      delete []label;
      zoom_help(st_file_id_,  path_file_id_, root_path_, root_label_, root_same_, root_other_, visited_count_);
    }
  }
  delete []bpath;
}

void zoom(int node_id, char *st_file_, char *path_file_, 
	  suffix_tree_node *&root_same_, suffix_tree_node *&root_other_) {
  int st_index_file_id, st_file_id, offset1, offset2;
  int clicked_node_path_len, label_len, path_file_id;
  char *clicked_node_path, *label;

  struct timeval startTime;
  struct timeval stopTime;

  path_file_id = open(path_file_, O_CREAT | O_WRONLY | O_TRUNC, S_IREAD | S_IWRITE);

  char *st_index_file = new char[strlen(st_file_) + strlen(".idx") + 1];
  sprintf(st_index_file, "%s.idx", st_file_);
  st_index_file_id = open(st_index_file, O_RDONLY);
  delete []st_index_file;

  if(lseek(st_index_file_id, node_id * 2 * sizeof(int), SEEK_SET) < 0) return;
  if(read(st_index_file_id, (char *)&offset1, sizeof(int)) < sizeof(int)) return;
  if(read(st_index_file_id, (char *)&offset2, sizeof(int)) < sizeof(int)) return;
  close(st_index_file_id);
  
  st_file_id = open(st_file_, O_RDONLY);
  if(lseek(st_file_id, offset1, SEEK_SET) < 0) return;
  if(read(st_file_id, (char *)&clicked_node_path_len, sizeof(int)) < sizeof(int)) return;
  clicked_node_path = new char[clicked_node_path_len + 1];
  if(read(st_file_id, (char *) clicked_node_path, clicked_node_path_len) < clicked_node_path_len) return;
  clicked_node_path[clicked_node_path_len] = 0;
 
  if(lseek(st_file_id, offset2, SEEK_SET) < 0) return;
  if(read(st_file_id, (char *)&label_len, sizeof(int)) < sizeof(int)) return;
  label = new char[label_len + 1];
  if(read(st_file_id, (char *)label, label_len) < label_len) return;
  label[label_len] = 0;

  int visited_count = 0;
  assert(gettimeofday(&startTime, NULL) != -1);
  zoom_help(st_file_id,  path_file_id, clicked_node_path, label, root_same_, root_other_, visited_count);
  assert(gettimeofday(&stopTime, NULL) != -1);
  double zoom_time = ((stopTime.tv_sec-startTime.tv_sec) + 
		      (stopTime.tv_usec-startTime.tv_usec) * 1e-6);
  StatisticsFileStream << visited_count << " "; // 1. Visited Node Count
  StatisticsFileStream << zoom_time << " "; // 2. Zoom Time

  /*
  root_same_ = new suffix_tree_node;
  root_same_->path = clicked_node_path;
  root_same_->is_leaf = false;
  root_same_->u.non_leaf.num_children = 1;
  root_same_->u.non_leaf.labels = new (char *)[1];
  root_same_->u.non_leaf.labels[0] = label;
  root_same_->u.non_leaf.children = new (suffix_tree_node *)[1];
  root_same_->u.non_leaf.children[0] = NULL;
  
  zoom_help(st_file_id, clicked_node_path, label, root_same_->u.non_leaf.children[0], root_other_, visited_count);
  */
  close(st_file_id);
  close(path_file_id);
}

void spaces(int level_) {
  for(int i=0; i<level_*4; i++) 
    cout << " ";
}

void print_text_suffix_tree(int level_, suffix_tree_node *root_) {
  if(root_) {
    spaces(level_);
    cout << "path = " << root_->path << endl;
    spaces(level_);
    cout << "is_leaf is " << (root_->is_leaf ? "true" : "false") << endl;
    if(root_->is_leaf == false) {
      spaces(level_);
      cout << "children = " << endl;
      for(int i=0; i<(root_->u.non_leaf.num_children); i++) {
	spaces(level_);
	if(strcmp(root_->u.non_leaf.labels[i], "") == 0) 
	  cout << "(eps)" << " [" << root_->u.non_leaf.num_docs[i] << "]" << endl;
	else 
	  cout << root_->u.non_leaf.labels[i] << " [" << root_->u.non_leaf.num_docs[i] << "]" << endl;
	print_text_suffix_tree(level_ + 1, root_->u.non_leaf.children[i]);	
      }
    }
    else {
      spaces(level_);
      cout << "paths = " << endl;
      for(int i=0; i<(root_->u.leaf.num_paths); i++) {
	spaces(level_);
	if(strcmp(root_->u.leaf.super_paths[i], "") == 0)
	  cout << "(eps) ";
	else
	  cout << root_->u.leaf.super_paths[i] << " ";
	cout << root_->u.leaf.file_offs[i] << " ";
	cout << root_->u.leaf.num_docs[i] << endl;
      }
    }
  }
}

void print_text(suffix_tree_node *root_) {
  print_text_suffix_tree(0, root_);
}

void insert_tree_node(tree_node *& tree_node_list_, int &count_, tree_node new_node_) {
  if(tree_node_list_ == NULL) {
    tree_node_list_ = new tree_node[1];
    tree_node_list_[0] = new_node_;
    count_ = 1;
  }
  else {
    count_++;
    tree_node *temp = tree_node_list_;
    tree_node_list_ = new (tree_node)[count_];
    for(int i=0; i<(count_-1); i++) 
      tree_node_list_[i] = temp[i];
    tree_node_list_[count_-1] = new_node_;
    delete []temp;
  }
  return;
}


void get_tree_stats(int level_, int print_flag_, int threshold_, int &node_id_, suffix_tree_node *root_, 
		    int &max_level_, int &count_, int &size_) {
  if(root_) {
    size_ += (sizeof(int)*3 + strlen(root_->path) + sizeof(bool) + sizeof(int)*4);
    cout << "Usage (so far) = " << size_ << endl;

    if(level_ > max_level_) max_level_ = level_;
    for(int i=0; i<(root_->u.non_leaf.num_children); i++) {
      size_ += (sizeof(int)*3 + strlen(root_->u.non_leaf.labels[i])+1);
      if(print_flag_) {
	count_++;
      }

      node_id_++;
      if(root_->u.non_leaf.children[i]->is_leaf == false) {
	if(print_flag_ == false || root_->u.non_leaf.num_docs[i] >= threshold_)
	  get_tree_stats(level_+1, print_flag_, threshold_, node_id_, root_->u.non_leaf.children[i], 
			 max_level_, count_, size_);
	else
	  get_tree_stats(level_+1, false, threshold_, node_id_, root_->u.non_leaf.children[i], 
			 max_level_, count_, size_);
      }      
    }
  }
}

void linearize_tree(int level_, int print_flag_, int threshold_, int &node_id_, suffix_tree_node *root_, 
		    int &max_level_, tree_node* &tree_node_list_, int &count_) {
  tree_node new_node;
  if(root_) {
    if(level_ > max_level_) max_level_ = level_;
    for(int i=0; i<(root_->u.non_leaf.num_children); i++) {
      if(print_flag_) {
	new_node.label = new char[strlen(root_->u.non_leaf.labels[i]) + 1];
	strcpy(new_node.label, root_->u.non_leaf.labels[i]);
	new_node.level = level_;
	new_node.is_leaf = root_->u.non_leaf.children[i]->is_leaf;
	new_node.num_docs = root_->u.non_leaf.num_docs[i];
	new_node.node_id = node_id_;
	insert_tree_node(tree_node_list_, count_, new_node);
      }

      node_id_++;
      if(root_->u.non_leaf.children[i]->is_leaf == false) {
	if(print_flag_ == false || root_->u.non_leaf.num_docs[i] >= threshold_)
	  linearize_tree(level_+1, print_flag_, threshold_, node_id_, root_->u.non_leaf.children[i], 
			 max_level_, tree_node_list_, count_);
	else
	  linearize_tree(level_+1, false, threshold_, node_id_, root_->u.non_leaf.children[i], 
			 max_level_, tree_node_list_, count_);
      }      
    }
  }
}


void print_html_trees(int view_, int threshold_, 
		      suffix_tree_node *root_sup_, suffix_tree_node *root_sub_, char *client_id,
		      int &sub_node_count, int &max_sub_level, 
		      int &sup_node_count, int &max_sup_level) {
  int node_id = 0, max_iterns;
  char *label_rev;
  tree_node *sub_tree_nodes = NULL;
  tree_node *sup_tree_nodes = NULL;

  node_id = 0;
  linearize_tree(0, true, threshold_, node_id, root_sub_, max_sub_level, sub_tree_nodes, sub_node_count);  

  node_id = 0;
  linearize_tree(0, true, threshold_, node_id, root_sup_, max_sup_level, sup_tree_nodes, sup_node_count);

  max_iterns = max(sub_node_count, sup_node_count);
  cout << "<table border=\"1\">" << endl;
  cout << "<tr align=\"center\">";

  if (max_sup_level >= 1) {
    cout << "<th colspan="
	 << "\"" << max_sup_level << "\">"
	 << "Super Tree"
	 << "</th>" << endl;
    
  }

  if (sup_node_count > 0) 
    cout << "<th colspan=\"1\"> Root Super </th>";
  
  if (sub_node_count > 0) 
    cout << "<th colspan=\"1\"> Root Sub </th>";
  
  if (max_sub_level >= 1) 
    cout << "<th colspan="
	 << "\"" << max_sub_level << "\">"
	 << "Sub Tree"
	 << "</th>" << endl;  
  cout << "</tr>" << endl << endl;
  
  for(int i=0; i<max_iterns; i++) {
    cout << "<tr>" << endl;
    for(int j=-max_sup_level; j<=max_sub_level; j++) {
      if(j <= 0 && sup_node_count > 0) {
	cout << "<td> ";          
	if(i < sup_node_count && -j == sup_tree_nodes[i].level) {
	  //if(sup_tree_nodes[i].is_leaf == false && sup_tree_nodes[i].num_docs < threshold_)
	  cout << "<a href = \""
	       << "http://mjolnir.cs.umd.edu:1975/cgi-bin/new_search?"
	       << "operation=zoom"
	       << "&database=0"
	       << "&node=" << sup_tree_nodes[i].node_id 
	       << "&view=" << view_
	       << "&id=" << client_id
	       << "\">"; // url	
	  //else 
	  if(strcmp(sup_tree_nodes[i].label, "") == 0)
	    cout << "(eps)";
	  else {
	    label_rev = reverse_path(sup_tree_nodes[i].label);
	    cout << label_rev;
	    delete []label_rev;
	  }
	  cout << "</a> ";
	  cout << "<a href = \""
	       << "http://mjolnir.cs.umd.edu:1975/cgi-bin/new_search?"
	       << "operation=getdocs"
	       << "&database=0"
	       << "&node=" << sup_tree_nodes[i].node_id 
	       << "&view=" << view_
	       << "&id=" << client_id
	       << "\">"; // url
	  cout  << " [" << sup_tree_nodes[i].num_docs << "] ";
	  cout << "</a> ";
	}
	else {
	  cout << "&nbsp;";
	}
	cout << "</td> ";          
      }
      if(j >= 0 && sub_node_count > 0) {
	cout << "<td> ";          
	if(i < sub_node_count && j == sub_tree_nodes[i].level) {
	  //if(sub_tree_nodes[i].is_leaf == false && sub_tree_nodes[i].num_docs < threshold_)
	  cout << "<a href = \""
	       << "http://mjolnir.cs.umd.edu:1975/cgi-bin/new_search?"
	       << "operation=zoom"
	       << "&database=1"
	       << "&node=" << sub_tree_nodes[i].node_id
	       << "&view=" << view_
	       << "&id=" << client_id
	       << "\">"; // url	
	  if(strcmp(sub_tree_nodes[i].label, "") == 0)
	    cout << "(eps)";
	  else 
	    cout << sub_tree_nodes[i].label;	  
	  cout << "</a> ";
	  //else 
	  cout << "<a href = \""
	       << "http://mjolnir.cs.umd.edu:1975/cgi-bin/new_search?"
	       << "operation=getdocs"
	       << "&database=1"
	       << "&node=" << sub_tree_nodes[i].node_id 
	       << "&view=" << view_
	       << "&id=" << client_id
	       << "\">"; // url
	  cout  << " [" << sub_tree_nodes[i].num_docs << "] ";
	  cout << "</a> ";
	}
	else {
	  cout << "&nbsp;";
	}
	cout << "</td>" << endl;
      }
    }  
    cout << "</tr>" << endl << endl;        
  }
  cout << "</table>" << endl;

  delete []sub_tree_nodes;
  delete []sup_tree_nodes;
}


//#define UNCOMMENT_MAIN
#ifdef UNCOMMENT_MAIN
main() {
  char path1[100];
  char super_path1[100];

  suffix_tree_node *root = NULL;//, *temp;

  while(!cin.eof()) {
    cout << "Enter path to insert: ";
    cin >> path1;
    if(cin.eof()) break;

    cin >> super_path1;
    if(strcmp(super_path1, "(epsilon)") == 0)
      insert_path(root, "", path1, "", 0, 0);
    else
      insert_path(root, "", path1, super_path1, 0, 0);
  }
  cout << endl;
  print_html(0, root);
  delete_suffix_tree(root);
}

#endif
