#include "Index5.h"
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>

ofstream WordListWriteStream;
ofstream RunGenerationStream;

extern "C" int errno;
Index::Index(char *fname_) {
  char *fname;
  ifstream in;
  int run;
  _fname = new char[strlen(fname_)+1];
  if(_fname == NULL) {
    return;
  }
  strcpy(_fname, fname_);
  
  _dictionary = new Trie();

  _run_number = 0;

  struct stat buf;
  fname = new char[strlen(_fname) + strlen(".run") + 1];
  sprintf(fname, "%s.run", _fname);
  if(stat(fname, &buf) == 0) {
    in.open(fname);
    in >> run;
    if(run >= _run_number) _run_number = run + 1;
    in.close();
  }
  else {
    _run_number = 0;
  }
  delete []fname;
}

Index::~Index() {
  char *fname;
  ofstream out;

  flush_dictionary();
  
  fname = new char[strlen(_fname) + strlen(".run") + 1];  
  sprintf(fname, "%s.run", _fname);
  out.open(fname);  
  out << _run_number << endl;
  out.close();
  delete fname;
  delete []_fname;
}


int Index::insert(char *key_, char *path_, int doc_id_, int word_pos_) {
  if(_dictionary == NULL) _dictionary = new Trie();
  if(_dictionary->insert(key_, path_, doc_id_, word_pos_) == -1) {
    return -1;
  }

  if(_dictionary->getTrieSize() >= MAX_TRIE_SIZE) {
    if(flush_dictionary() == -1) {
      return -1;
    }
  }
  return 0;
}

InvertedFile* Index::create_inverted_file(int run_) {
  char *fname = new char[strlen(_fname) + 1 + strlen("..if") + 10];
  sprintf(fname, "%s.%d.if", _fname, run_);
  //cout << "creating " << fname << endl;
  InvertedFile *ptr = new InvertedFile(fname, fname);
  delete []fname;
  return ptr;
}

int Index::flush_dictionary() {
  InvList inv_list;
  String word;
  List<String, String> *words;
  InvertedFile *inverted_file;
  double dict_find_time, run_genern_time, if_insert_time, dict_traverse_time, list_traverse_time;
  struct timeval startTime;
  struct timeval stopTime;
  int list_len, list_count, word_len, num_words_in_run = 0;

  if(_dictionary == NULL) return 0;


  struct timeval time_struct;
  assert(gettimeofday(&time_struct, NULL) != -1);

  // compute inv_list sizes
  _dictionary->compute_sizes();

  // write all inverted lists for this run
  assert(gettimeofday(&startTime, NULL) != -1);
  words = _dictionary->get_all_words();
  assert(gettimeofday(&stopTime, NULL) != -1);
  dict_traverse_time = 
    (stopTime.tv_sec-startTime.tv_sec) + 
    (stopTime.tv_usec-startTime.tv_usec) * 1e-6;

  if(words->isEmpty()) {
    delete words;
    delete _dictionary;
    _dictionary = NULL;
    return 0;
  }
  
  RunGenerationStream.open("../stats/run_gen_stats.log", ios::out | ios::app);
  WordListWriteStream.open("../stats/word_list_write_stats.log", ios::out | ios::app);

  run_genern_time = 0;

  assert(gettimeofday(&startTime, NULL) != -1);
  words->goFirst();
  // create new inverted file
  inverted_file = create_inverted_file(_run_number);

  while(!words->isOffList()) {
    // get word in dictionary
    word = words->getFromCurrent();
    assert(gettimeofday(&stopTime, NULL) != -1);
    list_traverse_time = 
      (stopTime.tv_sec-startTime.tv_sec) + 
      (stopTime.tv_usec-startTime.tv_usec) * 1e-6;
 
    word_len = word.get_len();

    // find inverted list for word
    assert(gettimeofday(&startTime, NULL) != -1);
    _dictionary->find(word, inv_list);
    assert(gettimeofday(&stopTime, NULL) != -1);

    dict_find_time = 
      (stopTime.tv_sec-startTime.tv_sec) + 
      (stopTime.tv_usec-startTime.tv_usec) * 1e-6;

    // add inverted list for inverted file
    assert(gettimeofday(&startTime, NULL) != -1);
    inverted_file->insert_record(word, inv_list, Time(time_struct.tv_sec, time_struct.tv_usec));
    assert(gettimeofday(&stopTime, NULL) != -1);

    if_insert_time = 
      (stopTime.tv_sec-startTime.tv_sec) + 
      (stopTime.tv_usec-startTime.tv_usec) * 1e-6;
 
    // print inverted list stats
    list_len = inv_list.get_print_len();
    list_count = inv_list.get_num_entries();
    WordListWriteStream << word_len << " " 
			<< dict_find_time << " " 
			<< list_len << " " 
			<< list_count << " " 
			<< if_insert_time << endl;

    run_genern_time += (dict_find_time + if_insert_time+list_traverse_time);
    num_words_in_run++;

    assert(gettimeofday(&startTime, NULL) != -1);
    words->delCurr();	
  }
  delete inverted_file;

  RunGenerationStream << num_words_in_run << " " 
		      << dict_traverse_time << " " 
		      << run_genern_time << endl;

  WordListWriteStream.close();
  RunGenerationStream.close();

  _run_number++;

  delete words;
  delete _dictionary;
  _dictionary = NULL;
  return 0;
}
