#include "Indexer.h"
#include "List.h"
#include "String.h"
#include "Int.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h> // sprintf
#include <strings.h>
#include <assert.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include "url_docid_record.h" // URL_SIZE
#include "Index5.h" // NUM_BUCKETS
#include <math.h> // pow
 
#define FILE_BUF_SIZE (1024 * 1024)

extern unsigned int hash(char *key, int key_len);

extern "C" int errno;

extern ofstream GLOBAL_logFileStream;
Index* GLOBAL_index;

extern ofstream statFileStream;

Stack Indexer::_elmntStk = Stack();
char* Indexer::_path = NULL;
XML_Parser Indexer::_parser = XML_Parser();

UrlDocMap* Indexer::_url_docid_map = 0;
int Indexer::_docid_url_map = -1;
char* Indexer::_fname = 0;

int Indexer::_docCount = -1;  
int Indexer::_wordCount = -1;

void Indexer::init(char *fname_) {
  _wordCount = 0;

  _fname = new char[strlen(fname_)+1];
  strcpy(_fname, fname_);

  char *fname = new char[strlen(_fname) + strlen(".start_doc")+1];  
  sprintf(fname, "%s.start_doc", _fname);
  ifstream in(fname);
  if(!in) 
    _docCount = 0;
  else {
    in >> _docCount;
    if(!in) _docCount = 0;
  }
  in.close();
  delete []fname;

  fname = new char[strlen(_fname) + strlen(".ids") + 1];
  sprintf(fname, "%s.ids", _fname);
  struct stat buf;
  if( (stat(fname, &buf) < 0) && (errno == ENOENT)) {
    _docid_url_map = open(fname, O_CREAT | O_WRONLY, S_IRWXU);
    GLOBAL_logFileStream << "index: Creating ids file; fd = " << _docid_url_map << endl;
  }
  else {
    _docid_url_map = open(fname, O_RDWR);
    GLOBAL_logFileStream << "index: Adding to ids file; fd = " << _docid_url_map << endl;
    if(_docid_url_map < 0)
      perror("Thomas");
  }
  delete []fname;
  
  fname = new char[strlen(_fname) + strlen(".urls") + 1];
  sprintf(fname, "%s.urls", _fname);
  _url_docid_map = new UrlDocMap(fname);
  delete []fname;
  
  GLOBAL_index = new Index(_fname);  
}

void Indexer::cleanup() {
  char *fname = new char[strlen(_fname) + strlen(".start_doc")+1];  
  sprintf(fname, "%s.start_doc", _fname);
  ofstream out(fname);
  if(out) 
    out << _docCount;
  else
    GLOBAL_logFileStream << "Index: " << "open of start_doc in cleanup failed" << endl;
  out.close();
  delete []fname;

  // close the ids file
  close(_docid_url_map);
  delete GLOBAL_index;
  delete _url_docid_map;
  delete []_fname;
}

void Indexer::startElement(void *userData, const char *name, const char **atts)
{
  bool ignoreTag = false;

  _elmntStk.push(name);

  if(_path != NULL) delete []_path;
  _path = (Indexer::_elmntStk).concat('.');
  if(ignoreTag == false) {
    int j=0;
    while(atts && atts[j]) {
      startElement(NULL, atts[j], NULL);
      charElement(NULL, atts[j+1], strlen(atts[j+1]));
      endElement(NULL, atts[j]);
      j+=2;
    }
  }
}

void Indexer::endElement(void *userData, const char *name)
{
  char* s;
  int fanout, node_size;

  if(_elmntStk.pop(s, fanout, node_size))
    delete []s;
  if(_path != NULL) delete []_path;
  _path = (Indexer::_elmntStk).concat('.');
}


void Indexer::charElement(void *userData, const char *chars, int txtLen)
{
  int i;
  char *buffer = new char[txtLen+1];
  char *token;

  // convert chunk of text to lowercase
  for(i=0; i<txtLen; i++)
    buffer[i] = tolower(chars[i]);
  buffer[txtLen] = 0;

  // find the first word in this chunk of text
  token = strtok(buffer, "\n \t!@#$%^&*()-_+=|\\~`:;<,>.?/:;\"'[{]}");

  while(token) {
    // strip punctuation characters from the word
    strip_affixes(token); 
      
    if(strlen(token) > 0) {
      // increment the word counter
      _wordCount++;

      assert(GLOBAL_index->insert(token, _path, _docCount, _wordCount) == 0);	  
    }

    // find the next word
    token = strtok(NULL, "\n \t!@#$%^&*()-_+=|\\~`:;<,>.?/:;\"'[{]}");
  }

  delete []buffer;
}


int Indexer::parseAndIndex(char *fname_, char *url_) 
{
  int file_len, file_desc, batch_size;
  char *url_string;
  
  char buf[FILE_BUF_SIZE];
  int done = 0, retval = 0, nbytes;

  // reset the word counter
  _wordCount  = 0;

  // write the id, url mapping
  if(strlen(url_) >= URL_SIZE) {
    retval = -1; 
    return retval;
  }

  url_string = new char[URL_SIZE+1];
  memset(url_string, ' ', URL_SIZE);
  url_string[URL_SIZE] = 0;
  strcpy(url_string, url_);

  if(lseek(_docid_url_map, _docCount * URL_SIZE, SEEK_SET) < 0) {
    delete []url_string;
    GLOBAL_logFileStream << "index: lseek in the ids file to offset " << _docCount * URL_SIZE << " failed" << endl;
    return -1;
  }

  nbytes = write(_docid_url_map, url_string, URL_SIZE);
  if(nbytes < URL_SIZE) { 
    delete []url_string;
    GLOBAL_logFileStream << "index: error writing docid->url mapping\n";
    return -1;
  }

  delete []url_string;

  double cpu_time, disk_time;
  if(_url_docid_map->insert(String(url_), Int(_docCount), cpu_time, disk_time) == -1) {
    GLOBAL_logFileStream << "index: error writing url->docid mapping\n";
    retval = -1;
    assert(false);
  }

  // increment the document counter
  _docCount++; 
  
  // index the file  
  struct stat stat_buf;
  if(stat(fname_, &stat_buf) == 0) {
    file_len = stat_buf.st_size;
    
    file_desc = open(fname_, O_RDONLY);
    if(file_desc < 0) {
      perror("Thomas:");
      GLOBAL_logFileStream << "index: error opening file " << fname_ << endl;
      return -1;
    }

    // create and init the parser
    XML_Parser parser = XML_ParserCreate(NULL);
    XML_SetUserData(parser, NULL);
    XML_SetElementHandler(parser, startElement, endElement);
    XML_SetCharacterDataHandler(parser, charElement);
    
    batch_size = FILE_BUF_SIZE;
    while(file_len >= batch_size) {
      if(read(file_desc, buf + FILE_BUF_SIZE - batch_size, batch_size) < batch_size) {
	GLOBAL_logFileStream << "index: error reading from file " << fname_ << endl;
	retval = -1;
	goto END;
      }
      
      batch_size = FILE_BUF_SIZE;
      while(!isspace(buf[batch_size-1])) batch_size--;

      if(!XML_Parse(parser, buf, batch_size, done = 0)) {
	statFileStream << "index: " << XML_ErrorString(XML_GetErrorCode(parser)) << endl;
	_elmntStk.empty();
	retval = -1;
	goto END;
      }

      char *temp = new char[FILE_BUF_SIZE-batch_size];
      memcpy(temp, buf + batch_size, FILE_BUF_SIZE-batch_size);
      memcpy(buf, temp, FILE_BUF_SIZE-batch_size);
      delete []temp;

      file_len -= batch_size;
    }
    
    if(read(file_desc, buf + FILE_BUF_SIZE - batch_size, file_len) < file_len) {
      GLOBAL_logFileStream << "index: error reading from file " << fname_ << endl;
      retval = -1;
      goto END;
    }
    buf[FILE_BUF_SIZE - batch_size + file_len] = '\n';
    if(!XML_Parse(parser, buf, FILE_BUF_SIZE - batch_size + file_len + 1, done = 0)) {
      statFileStream << "index: " << XML_ErrorString(XML_GetErrorCode(parser)) << endl;
      _elmntStk.empty();
      retval = -1;
      goto END;
    }


  END:
    // clean-up
    XML_ParserFree(parser);
    
    close(file_desc);
  }

  return retval;
}

