#include "Stack.h"

Stack::Stack(int maxSize) {
  count = 0;
  size = maxSize;
  array = new (char*)[size];
  fanout = new int[size];
  nodesize = new int[size];
  for(int i=0; i < size; i++) {
    fanout[i] = 0;
    nodesize[i] = 0;
  }
}

Stack::Stack(const Stack &copy) {
  count = copy.count;
  size = copy.size;
  array = new (char *)[size];
  fanout = new int[size];
  nodesize = new int[size];
  for(int i=0; i<count; i++) {
    int len  = strlen(copy.array[i])+1;
    array[i] = new char[len];    
    memcpy(array[i], copy.array[i], len);
    fanout[i] = copy.fanout[i];
    nodesize[i] = copy.nodesize[i];
  }
}

void Stack::empty() {
  for(int i=0; i<count; i++) {
    delete [](array[i]);
    fanout[i] = 0;
    nodesize[i] = 0;
  }
  count = 0;

}  

Stack::~Stack() {
  for(int i=0; i<count; i++) 
    delete [](array[i]);
  delete []array;
  delete []fanout;
  delete []nodesize;
}

bool Stack::push(const char* d) {
  if (count == size) return false;
  else {
    int len      = strlen(d) + 1;
    array[count] = new char[len];    
    memcpy(array[count], d, len);
    if(count > 0)
      fanout[count - 1]++;
    count++;
    return true;
  }
}

bool Stack::pop(char* &d, int &fan_out, int &node_size) {
  if(count == 0) return false;
  else {
    count --;
    int len = strlen(array[count]) + 1;
    d = new char[len];
    memcpy(d, array[count], len);    
    fan_out = fanout[count];
    node_size = nodesize[count];
    fanout[count] = 0;
    nodesize[count] = 0;
    delete [](array[count]);
    return true;
  }
}

void Stack::addTokenCount(int token_count_) {
  if(count > 0)
    nodesize[count-1] += token_count_;
}

char* Stack::concat(char separator) {
  int  len = 0;
  char *buf;

  for(int i=0; i<count; i++) 
    len += strlen(array[i]);

  buf = new char[len + count];
  char *ptr = buf;
  for(int i=0; i<count; i++) { 
    int leni  = strlen(array[i]);
    memcpy(ptr, array[i], leni);
    ptr      += leni;
    ptr[0]    = separator;
    ptr++;
  }
  ptr[-1]      = 0;

  return buf;
}
