#include "String.h"
#include <string.h>

#define MAX_STRING_LEN 200
int String::max = MAX_STRING_LEN;

// create and initialize a String object. 
// the default value for the String is an empty String.
String::String(char *init) 
{
  data = NULL;

  // call the assignment operator to do the job for you
  *this = init;
}

// copy constructor for String class
String::String(const String& copyFrom)
{
  data = NULL;

  // call the assignment operator to do the copying
  *this = copyFrom;
}

// destructor for String class
String::~String()
{
  free();
}


char *String::toCharPtr() {
  char *ptr = new char[size];
  strcpy(ptr, data);
  return ptr;
}

// assignment operator
String& String::operator=(const String& assignFrom)
{
  int counter;

  // free the existing data array
  if(data != NULL) free();

  size = assignFrom.size;

  // allocate array for holding the String
  data = new char[size];
  
  counter = 0;
  while(counter < size)
    {
      data[counter] = assignFrom.data[counter];
      counter++;
    }
  
  return *this;
}


// assignment operator
String& String::operator = (const char * init)
{
  int counter;
  
  if(data != NULL) free();

  // initialize length of String to 0
  size = 0;

  // assume that init is not NULL
  // find the length of the argument String
  while(init[size] != '\0')
    {
      size++;
    }
  
  // allocate room for holding the String
  size++;

  data = new char[size];
  
  // copy characters from init to data  
  counter = 0;
  while(counter < size)
    {
      data[counter] = init[counter];
      counter++;
    }

  return *this;
}


// "precedes" operator
// returns 1 if current String precedes other String
// returns 0 otherwise
int String::operator < (const String &other) const
{
  int counter = 0;

  while(counter < size && counter < other.size)
    {
      if(data[counter] < other.data[counter]) return 1;
      if(data[counter] > other.data[counter]) return 0;
      counter ++;
    }
  
  if(size < other.size) return 1;
  return 0;
}

// "succeeds" operator
// returns 1 if current String succeeds other String
// returns 0 otherwise 
int String::operator > (const String &other) const
{
  int counter = 0;

  while(counter < size && counter < other.size)
    {
      if(data[counter] < other.data[counter]) return 0;
      if(data[counter] > other.data[counter]) return 1;
      counter ++;
    }
  
  if(size > other.size) return 1;
  return 0;
}

// "equal" operator
// returns 1 if current String is the same as the other String
// returns 0 otherwise
int String::operator == (const String &other) const
{
  int counter = 0;

  while(counter < size && counter < other.size)
    {
      if(data[counter] < other.data[counter]) return 0;
      if(data[counter] > other.data[counter]) return 0;
      counter ++;
    }
  
  if(size == other.size) return 1;
  return 0;
}

int String::operator <= (const String &other) const
{
  return (*this < other || *this == other);
}

int String::operator >= (const String &other) const
{
  return (*this > other || *this == other);
}

// input operator
istream& operator >> (istream &in, String &s)
{
  char nextChar;
  char* array = new char[String::max];

  if (s.data != NULL) s.free();

  // discard the leading whitespace characters
  in >> ws;
  nextChar = in.peek();
  if(nextChar != '"') {
    // read until a whitespace character is encountered
    s.size = 0;
    while((s.size < String::max) && 
	  (nextChar = in.get()) &&
	  !in.eof() && 
	  nextChar != ' ' && 
	  nextChar != '\t' &&
	  nextChar != '\n')
      {
	array[s.size] = nextChar;
	s.size++;
      }
  }
  else {
    nextChar = in.get();
    // read until a closing " is encountered
    s.size = 0;
    while((s.size < String::max) && ((nextChar = in.get()) != '"'))
      {
	array[s.size] = nextChar;
	s.size++;
      }
  }

  array[s.size++] = 0;

  // copy characters from array into s.data
  s.data = new char[s.size];
  for(int i=0; i<s.size; i++)
    s.data[i]=array[i];

  delete []array;
  return in;
}

// output operator
ostream& operator << (ostream &out, const String s)
{
  //int counter;
  out << s.data;
  return out;
}

// free String
void String::free()
{
  delete [] data;
}

/*

// test function for String classs
void main()
{
  String A("hello");
  String B("my name is");
  String C = "Thomas Baby";
  String P = " ";
  String N = "\n";

  String L=" < ";
  String G=" > ";
  String E=" == ";

  cout << A << L << B << P << (A<B) << N;
  cout << A << G << B << P << (A>B) << N;
  cout << A << E << B << P << (A == B) << N;

  cout << A << P << B << P << C << N ;
  String myString;
  cin >> myString;
  cout << "myString = \"" << myString << "\"" << endl;
}


*/
