#include "Trie4.h"
#include <assert.h>
#include <fstream.h>

extern ofstream GLOBAL_logFileStream;

void Trie::freeHelp(TrieNode *node_ptr_) {

  if(node_ptr_->list)
    delete node_ptr_->list;

  TrieNode *temp = node_ptr_->child;
  if(temp)
    this->freeHelp(temp);

  temp = node_ptr_->next;
  if(temp)
    this->freeHelp(temp);
  
  delete node_ptr_;
}


void Trie::get_all_words_helper(char* word, TrieNode *node_ptr_, List<String, String> &list_) {
  int len = strlen(word);

  char *store = new char[len + 1];
  memcpy(store, word, len+1);

  char *buf   = new char[len + 2];
  memcpy(buf, word, len);
  buf[len] = node_ptr_->letter;
  buf[len + 1] = 0;

  if(node_ptr_->list && !(node_ptr_->list->isEmpty())) {
    list_.addToBack(String(buf));
  }

  TrieNode *temp = node_ptr_->child;
  if(temp)
    this->get_all_words_helper(buf, temp, list_);

  temp = node_ptr_->next;
  if(temp)
    this->get_all_words_helper(store, temp, list_);
  delete []buf;
  delete []store;
}

void Trie::compute_sizes_helper(TrieNode *node_ptr_) {
  if(node_ptr_->list && !(node_ptr_->list->isEmpty())) {
    node_ptr_->size = node_ptr_->list->get_print_len();
  }
  else
    node_ptr_->size = 0;

  TrieNode *temp = node_ptr_->child;
  if(temp)
    this->compute_sizes_helper(temp);

  temp = node_ptr_->next;
  if(temp)
    this->compute_sizes_helper(temp);
}

// default constructor
Trie::Trie() {
  _root = 0;
  _num_entries = 0;
}

// destructor
Trie::~Trie() {
  if(_root) freeHelp(_root);
  _root = 0;
  _num_entries = 0;
}

// insert function
// adds a word, its path and its location to the trie
// returns -1 if there is an error
// returns 0 in all other cases
int Trie::insert(String word_, String pname_, int doc_id_, int word_posn_) {
  char *word = word_.toCharPtr();
  char *pname = pname_.toCharPtr();

  if(word == 0) {
    delete []word;
    delete []pname;
    return -1;
  }
  int i = 0;
  TrieNode *parent = 0;
  TrieNode *ptr = _root;
  TrieNode *prev = 0;
  while(word[i]) {
    while(ptr && ptr->letter < word[i]) {
      prev = ptr;
      ptr = ptr->next;
    }
    
    if(ptr == 0 || ptr->letter > word[i]) {
      TrieNode *temp = new TrieNode;
      temp->letter = word[i];
      temp->list = new InvList();
      temp->size = 0;
      temp->next = ptr;
      temp->child = 0;
      if(prev)
	prev->next = temp;
      else if(parent)
	parent->child = temp;
      else
	_root = temp;
      ptr  =  temp;
    }

    parent = ptr;
    if(word[i+1])
      ptr    = parent->child;
    prev   = 0;
    
    i++;
  }

  if(ptr == NULL) assert(0);
  if(ptr->list == NULL) ptr->list = new InvList();
  ptr->list->insert(pname, doc_id_, word_posn_);

  _num_entries++;

  delete []word;
  delete []pname;

  return 0;
}

void Trie::compute_sizes() {
   if(_root) compute_sizes_helper(_root);
}

bool Trie::find(String word_, InvList& list_) {
  char *word = word_.toCharPtr();
  
  if(word == 0) {
    delete []word;
    return false;
  }
  int i = 0;
  TrieNode *parent = 0;
  TrieNode *ptr = _root;
  TrieNode *prev = 0;
  while(word[i]) {
    while(ptr && ptr->letter < word[i]) {
      prev = ptr;
      ptr = ptr->next;
    }
    
    if(ptr == 0 || ptr->letter > word[i]) 
      return false;
    
    parent = ptr;
    if(word[i+1])
      ptr    = parent->child;
    prev   = 0;
    
    i++;
  }
  
  if(ptr == NULL) assert(0);
  if(ptr->list == NULL) ptr->list = new InvList();
  list_ = *(ptr->list);
  
  delete []word;
  
  return true;
}

List<String, String>* Trie::get_all_words() {
  List<String, String> *result = new List<String, String>();
  if(result == NULL) {
    GLOBAL_logFileStream << "Trie: " << "get_words; Unable to create list" << endl;
    exit(0);
  }
   if(_root) get_all_words_helper("", _root, *result);
  return result;
}


int Trie::getTrieSize() const {
  return _num_entries;
}

int Trie::get_size(String word_) {
  char *word = word_.toCharPtr();

  if(word == 0) {
    delete []word;
    return -1;
  }
  int i = 0;
  TrieNode *parent = 0;
  TrieNode *ptr = _root;
  TrieNode *prev = 0;
  while(word[i]) {
    while(ptr && ptr->letter < word[i]) {
      prev = ptr;
      ptr = ptr->next;
    }
    
    if(ptr == 0 || ptr->letter > word[i]) {
      assert(0);
    }

    parent = ptr;
    if(word[i+1])
      ptr    = parent->child;
    prev   = 0;
    
    i++;
  }

  delete []word;
  return ptr->size;
}
