#include "regex.h"

//#define TEST_REGULAR_EXPRN
#ifdef TEST_REGULAR_EXPRN
//////////////////////////////////////////////////////
//This block of code is for regular expression support
#define INIT         register char *sp=instring;
#define GETC()       (*sp++)
#define PEEKC()      (*sp)
#define UNGETC(c)     (--sp)
#define RETURN(c)    return 0;
#define ERROR(c)     exit(0);
#include <regexp.h>

char *compile(char *instring, char *expbuf,
	      const char *endbuf, int eof);

int step(const char *string, const char *expbuf);

int advance(const char *string, const char *expbuf);

extern char *loc1, *loc2, *locs;
//////////////////////////////////////////////////////
#endif

// createRegExp
// Converts a single, valid path expression and converts it into a 
// regular expression of the form expected by the regexp library.

// Valid path expressions include ?.a.b, +.a.b, a.b

// Args: path (I)      : path name 
//       regex(O)      : buffer to hold the output regular expression
//       outbufsize(I) : size of the buffer 'regex' in bytes

// Returns: -1 : if the output regular expression cannot fit in regex
//          0  : if successful

int createRegExp(char *path, char *regex, int outbufsize) {
  char *output = regex;
  char *token  = strtok(path, ".");
  int   maxlen = outbufsize;
  int   inclen;

  memset(output, maxlen, '\0');
  output[0] = '^';
  output++;

  while(token) {
    
    if(strcmp(token, "+") == 0) {
      inclen = strlen("[0-9a-zA-Z_.+-][0-9a-zA-Z_.+-]*");
      if(inclen < maxlen)
	strcpy(output, "[0-9a-zA-Z_.+-][0-9a-zA-Z_.+-]*");
      else 
	return -1;
    }
    else if(strcmp(token, "?") == 0) {
      inclen = strlen("[0-9a-zA-Z_+-][0-9a-zA-Z_+-]*");
      if(inclen < maxlen)
	strcpy(output, "[0-9a-zA-Z_+-][0-9a-zA-Z_+-]*");
      else 
	return -1;
    }
    else {
      inclen = strlen(token);
      if(inclen < maxlen)
	strcpy(output, token);
      else 
	return -1;
    }

    output += inclen;
    maxlen -= inclen;
    if(maxlen < 0)
      return -1;

    inclen = 2;
    if(inclen < maxlen) {
      output[0] = '\\'; 
      output[1] = '.';
    }
    else 
      return -1;

    output += inclen;
    maxlen -= inclen;
    if(maxlen < 0)
      return -1;
    
    token = strtok(NULL, ".");
  }
  
  // Removing the last '\.'
  if(output[-1] == '.') {
    output -= 2;
    output[0] = '$';
    output[1] = '\0';
  }
  else 
    return -1;

  return 0;
}

#ifdef TEST_REGULAR_EXPRN
int main(int argc, char *argv[]) {
  char expbuf[ESIZE];
  char regString[ESIZE];
  char inpString[ESIZE];
  char path[ESIZE];

  cin >> path;
  if(createRegExp(path, regString, ESIZE) == -1) {
    cout << path << ": Buffer size for regex exceeded in createRegExp" << endl;
    return 1;
  }
  cout << regString << endl;

  compile(regString, expbuf, &expbuf[ESIZE], '\0');
  while(!cin.eof()) {
    cin >> inpString;
    cout << inpString;
    if(step(inpString, expbuf))
      cout << ":\t Match" << endl;
    else
      cout << ":\t No Match" << endl;
  }

  return 0;
}
#endif


