#include "LinearHash3.h"
#include <strings.h>   // strcpy, strlen
#include <sys/types.h> // stat, lseek, open
#include <sys/stat.h>  // stat, open
#include <stdio.h>     // sprintf
#include <assert.h>    // assert
#include <unistd.h>    // lseek, read, write
#include <fcntl.h>     // open
#include <errno.h>     // ENOENT
#include <fstream.h>
#include <stdlib.h>
#include <math.h> // pow

#define FILL_FACTOR (0.01)

extern int hash(char *, int);
extern ofstream GLOBAL_logFileStream;
static bool GLOBAL_log = true;
extern int errno;

typedef struct BktHeader_ {
  int _free_slot;
  int _over_flow;
} BktHeader;

template <class key_type, class data_type, class record_type>
int LinearHash<key_type, data_type, record_type>::find_bkt(char *key_) {
  int hash_val = hash(key_, strlen(key_)) % _fheader._num_bkts;
  return hash_val;
}

template <class key_type, class data_type, class record_type>
int LinearHash<key_type, data_type, record_type>::goto_bkt(int bktnum_) {
  if(lseek(_file_desc, 
	   sizeof(HashFileHeader) + bktnum_ * (sizeof(BktHeader) + _fheader._num_slots * _fheader._recd_size),
	   SEEK_SET) < 0) return -1;
  return 0;
}

template <class key_type, class data_type, class record_type>
int LinearHash<key_type, data_type, record_type>::my_write(int fd_, char *buf_, int nbytes_) {
  return write(fd_, buf_, nbytes_);
}

template <class key_type, class data_type, class record_type>
int LinearHash<key_type, data_type, record_type>::my_read(int fd_, char *buf_, int nbytes_) {
  return read(fd_, buf_, nbytes_);
}

template <class key_type, class data_type, class record_type>
LinearHash<key_type, data_type, record_type>::LinearHash(char *file_name_, int num_bkts_, int num_slots_, int rec_len_, char *idString_ = "LH3: ") {
  int nbytes, count, num_bkts, bkt_len;
  char *bkt_buffer, *junk_ptr;
  BktHeader bkt_hdr;
  struct stat buf;
  
  _idString = new char[strlen(idString_) + 1];
  strcpy(_idString, idString_);

  _file_name = new char[strlen(file_name_) + 1];
  if(NULL == _file_name) {
    GLOBAL_logFileStream << _idString << ": constr; malloc failed" << endl;
    return;
  }
  strcpy(_file_name, file_name_);
  
  if( (stat(_file_name, &buf) < 0) && (errno == ENOENT) ) {
    _file_desc = open(_file_name, O_RDWR | O_CREAT, S_IREAD|S_IWRITE);
    if(_file_desc < 0) {
      GLOBAL_logFileStream << _idString << ": constr; file create failed" << endl;
      return;
    }

    // create a new file header
    _fheader._num_slots = num_slots_;
    _fheader._num_bkts = num_bkts_;
    _fheader._recd_size = 2*sizeof(int) + rec_len_;
    
    // write the new file header
    nbytes = my_write(_file_desc, (char *)&_fheader, sizeof(HashFileHeader));
    if(nbytes < sizeof(HashFileHeader)) {
      GLOBAL_logFileStream << _idString << ": constr; file hdr write failed" << endl;
      return;
    }

    bkt_hdr._over_flow = -1;
    bkt_hdr._free_slot = 0;

    bkt_len = _fheader._num_slots * _fheader._recd_size + sizeof(BktHeader);
    bkt_buffer = new char[bkt_len];
    junk_ptr = bkt_buffer;

    memcpy(junk_ptr, (char *)&bkt_hdr, sizeof(BktHeader));
    junk_ptr += sizeof(BktHeader);

    memset(junk_ptr, '0', bkt_len-sizeof(BktHeader));
    
    num_bkts = _fheader._num_bkts;

    count = 0;
    while(count < num_bkts) {
      if(my_write(_file_desc, (char *)bkt_buffer, bkt_len) < bkt_len) {
	GLOBAL_logFileStream << _idString << ": constr; bkt write failed" << endl;
	return;
      }

      count++;
    }

    delete []bkt_buffer;
    close(_file_desc);

    //if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": constr; created a new file" << endl;
  }
  
  _file_desc = open(_file_name, O_RDWR);
  if(_file_desc < 0) {
    GLOBAL_logFileStream << _idString << ": constr; file open failed" << endl;
    return;
  }

  nbytes = my_read(_file_desc, (char *)&_fheader, sizeof(HashFileHeader));
  if(nbytes < sizeof(HashFileHeader)) {
    GLOBAL_logFileStream << _idString << ": constr; file hdr read failed" << endl;
    return;
  }

  num_bkts = _fheader._num_bkts;
  _lists = new (List<record_type, key_type> *) [num_bkts];
  for(count = 0; count < num_bkts; count++) 
    _lists[count] = new List<record_type, key_type> ();
  _occupancy = 0;

  _temp = NULL;

  cout << _idString 
       << ": num_bkts = " << _fheader._num_bkts 
       << ", num_slots = " << _fheader._num_slots 
       << ", rec_size = " << _fheader._recd_size << endl;
}

template <class key_type, class data_type, class record_type>
LinearHash<key_type, data_type, record_type>::LinearHash(const LinearHash& other_) {
  int nbytes, count, num_bkts;
  struct stat buf;
 
  _idString = new char[strlen(other_._idString) + 1];
  strcpy(_idString, other_._idString);

  _file_name = new char[strlen(other_._file_name) + 1];
  if(NULL == _file_name) {
    GLOBAL_logFileStream << _idString << "copy constr; malloc failed" << endl;
    return;
  }
  strcpy(_file_name, other_._file_name);

  if( (stat(_file_name, &buf) < 0) && (errno == ENOENT) ) {
    GLOBAL_logFileStream << _idString << "copy constr; file " << _file_name << " does not exist" << endl;
    return;
  }

  _file_desc = open(_file_name, O_RDWR);
  if(_file_desc < 0) {
    GLOBAL_logFileStream << _idString << "copy constr; file open failed" << endl;
    return;
  }

  nbytes = my_read(_file_desc, (char *)&_fheader, sizeof(HashFileHeader));
  if(nbytes < sizeof(HashFileHeader)) {
    GLOBAL_logFileStream << _idString << "copy constr; file hdr read failed" << endl;
    return;
  }

  num_bkts = _fheader._num_bkts;
  _lists = new (List<record_type, key_type> *) [num_bkts];
  for(count = 0; count < num_bkts; count++) 
    _lists[count] = new List<record_type, key_type> ();
  _occupancy = 0;

  _temp = NULL;
}
  
template <class key_type, class data_type, class record_type>
LinearHash<key_type, data_type, record_type>::~LinearHash() {
  int nbytes, count, num_bkts;

  count = 0;
  GLOBAL_logFileStream << _idString << ": destr; flush to disk, count = " << ++count << endl;
  while(flush_to_disk())
    GLOBAL_logFileStream << _idString << ": destr; flush to disk, count = " << ++count << endl;

  num_bkts = _fheader._num_bkts;
  for(count = 0; count < num_bkts; count++) 
    delete _lists[count];
  delete []_lists;

  if(lseek(_file_desc, 0, SEEK_SET) < 0) {
    GLOBAL_logFileStream << _idString << ": destr; file seek to beginning failed" << endl;
    return;
  }
  
  nbytes = my_write(_file_desc, (char *)&_fheader, sizeof(HashFileHeader));
  if(nbytes < sizeof(HashFileHeader)) {
    GLOBAL_logFileStream << _idString << ": destr; file hdr write failed" << endl;
    return;
  }

  close(_file_desc);
  delete []_file_name;
  delete []_idString;
}

template <class key_type, class data_type, class record_type>
int LinearHash<key_type, data_type, record_type>::find(key_type key_, data_type &data_, bool &found_) {
  record_type record;
  key_type key;
  key_type key_inlist;
  int key_len, bkt, count, key_onfile_len, data_onfile_len, key_inlist_len, bkt_len;
  char *key_ptr, *key_onfile_ptr, *data_onfile_ptr, *key_inlist_ptr, *bkt_buffer, *tmp_bkt_ptr;
  BktHeader bkt_hdr;

  found_ = false;

  key_len = key_.get_len();
  key_ptr = key_.toCharPtr();

  bkt = find_bkt(key_ptr);

  //if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": find;(" << key_ptr << "), bkt = " << bkt << endl;

  // look in _lists first
  _lists[bkt]->goFirst();
  while(!_lists[bkt]->isOffList()) {
    record = _lists[bkt]->getFromCurrent();

    key_inlist = record.get_key();
    key_inlist_ptr = key_inlist.toCharPtr();
    key_inlist_len = key_inlist.get_len();

    if(key_inlist_len == key_len && memcmp(key_inlist_ptr, key_ptr, key_len) == 0) {
      if(!found_) found_ = true;
      data_ = record.get_data();
    }
    delete []key_inlist_ptr;

    _lists[bkt]->goNext();
  }

  if(found_) {
    //GLOBAL_logFileStream << _idString << ": find; " << key_ << " found in _lists[" << bkt << "] itself" << endl;
    delete []key_ptr;
    return 0;
  }

  // lseek to the bucket
  if(goto_bkt(bkt) < 0) {
    GLOBAL_logFileStream << _idString << ": find; lseek to bkt failed" << endl;
    delete []key_ptr;
    return -1;
  }

  bkt_len  = sizeof(BktHeader) + _fheader._num_slots * _fheader._recd_size;
  bkt_buffer = new char[bkt_len];

  // read the bucket 
  if(my_read(_file_desc, (char *)bkt_buffer, bkt_len) < bkt_len) {
    GLOBAL_logFileStream << _idString << ": find; read of bkt failed" << endl;
    delete []key_ptr;
    delete []bkt_buffer;
    return -1;
  }

  memcpy((char *)&bkt_hdr, bkt_buffer, sizeof(BktHeader));
  tmp_bkt_ptr = bkt_buffer + sizeof(BktHeader);

  // check the slots in the bucket
  count = 0;
  //if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": find; bkt_hdr._free_slot = " << bkt_hdr._free_slot << endl;

  while(!found_ && count < bkt_hdr._free_slot) {    

    memcpy((char *)&key_onfile_len, tmp_bkt_ptr, sizeof(int));
    tmp_bkt_ptr += sizeof(int);

    key_onfile_ptr = new char[key_onfile_len + 1];
    memcpy((char *)key_onfile_ptr, tmp_bkt_ptr, key_onfile_len);
    key_onfile_ptr[key_onfile_len] = 0;
    tmp_bkt_ptr += key_onfile_len;

    if(key_len == key_onfile_len && memcmp(key_ptr, key_onfile_ptr, key_len) == 0) {
      //if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": find; found key in bkt" << endl;
      found_ = true;
    }

    delete []key_onfile_ptr;

    if(found_ == true) {
      //if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": find; " << key_ << " found in [" << bkt << "." << count+1 << "]" << endl;

      memcpy((char *)&data_onfile_len, tmp_bkt_ptr, sizeof(int));
      tmp_bkt_ptr += sizeof(int);
      
      data_onfile_ptr = new char[data_onfile_len];
      memcpy((char *)data_onfile_ptr, tmp_bkt_ptr, data_onfile_len);
      tmp_bkt_ptr += data_onfile_len;

      data_ = data_type(data_onfile_ptr);
      
      delete []data_onfile_ptr;
    
      tmp_bkt_ptr += (_fheader._recd_size-2*sizeof(int)-key_onfile_len-data_onfile_len);
    }
    else {
      tmp_bkt_ptr += (_fheader._recd_size-sizeof(int)-key_onfile_len);
    }
    count++;
  }
  delete []bkt_buffer;
  delete []key_ptr;
  return 0;
}

template <class key_type, class data_type, class record_type>
int LinearHash<key_type, data_type, record_type>::remove(key_type key_) {
  bool found;
  key_type key;
  record_type record;
  char *key_ptr, *key_onfile_ptr, *bkt_buffer, *last_rec_ptr, *tmp_bkt_ptr;
  int bkt_no, key_onfile_len, count, key_len, bkt_len;
  BktHeader bkt_hdr;

  key_ptr = key_.toCharPtr();
  key_len = key_.get_len();

  // lseek to the bucket
  bkt_no = find_bkt(key_ptr);

  //if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": remove; (" << key_ptr << "), bkt = " << bkt_no << endl;

  // look in _lists first
  found = false;
  _lists[bkt_no]->goFirst();
  while(!_lists[bkt_no]->isOffList()) {
    record = _lists[bkt_no]->getFromCurrent();
    key = record.get_key();
    if(key == key_) {
      _lists[bkt_no]->delCurr();
      found = true;
      _occupancy--;
      //if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": remove; " << key_ << " found in _lists[" << bkt_no << "] itself" << endl;
    }
    else {
      _lists[bkt_no]->goNext();
    }
  }


  if(goto_bkt(bkt_no) < 0) {
    GLOBAL_logFileStream << _idString << ": remove; seek to bucket failed, bkt = " << bkt_no << endl;
    delete []key_ptr;
    return -1;
  }

  bkt_len  = sizeof(BktHeader) + _fheader._num_slots * _fheader._recd_size;
  bkt_buffer = new char[bkt_len];

  // read the bucket 
  if(my_read(_file_desc, (char *)bkt_buffer, bkt_len) < bkt_len) {
    GLOBAL_logFileStream << _idString << ": find; read of bkt failed" << endl;
    delete []key_ptr;
    delete []bkt_buffer;
    return -1;
  }
 
  memcpy((char *)&bkt_hdr, bkt_buffer, sizeof(BktHeader));
  tmp_bkt_ptr = bkt_buffer + sizeof(BktHeader);

  // check the slots in the bucket
  count = 0;
  found = false;
  while(!found && count < bkt_hdr._free_slot) {
    memcpy((char *)&key_onfile_len, tmp_bkt_ptr, sizeof(int));
    tmp_bkt_ptr += sizeof(int);

    key_onfile_ptr = new char[key_onfile_len + 1];
    memcpy((char *)key_onfile_ptr, tmp_bkt_ptr, key_onfile_len);
    key_onfile_ptr[key_onfile_len] = 0;
    tmp_bkt_ptr += key_onfile_len;

    if(key_len == key_onfile_len && memcmp(key_ptr, key_onfile_ptr, key_len) == 0) {
      //if(GLOBAL_log) GLOBAL_logFileStream << _idString << ": remove; " << key_ << " found in [" << bkt_no << "." << count+1 << "]" << endl;

      found = true;
      
      tmp_bkt_ptr -= (key_onfile_len + sizeof(int));
      last_rec_ptr = bkt_buffer + sizeof(BktHeader) + (bkt_hdr._free_slot-1)*_fheader._recd_size;

      memcpy(tmp_bkt_ptr, last_rec_ptr, _fheader._recd_size);
      tmp_bkt_ptr += (_fheader._recd_size);

      (bkt_hdr._free_slot)--;
    }
    else {
      tmp_bkt_ptr += (_fheader._recd_size-sizeof(int)-key_onfile_len);
    }
    delete []key_onfile_ptr;
    
    count++;
  }

  if(found) {
    memcpy(bkt_buffer, (char *)&bkt_hdr, sizeof(BktHeader));

    if(goto_bkt(bkt_no) < 0) {
      GLOBAL_logFileStream << _idString << ": remove; seek to bucket failed, bkt = " << bkt_no << endl;
      delete []key_ptr;
      delete []bkt_buffer;
      return -1;
    }

    if(my_write(_file_desc, (char *)bkt_buffer, bkt_len) < bkt_len) {
      GLOBAL_logFileStream << _idString << ": remove; write of bkt failed" << endl;
      delete []key_ptr;
      delete []bkt_buffer;
      return -1;
    }
  }
  delete []key_ptr;
  delete []bkt_buffer;

  return 0;
}

template <class key_type, class data_type, class record_type>
int LinearHash<key_type, data_type, record_type>::insert(record_type record_) {
  key_type key;
  char *key_ptr;
  int bkt_no;

  key = record_.get_key();
  key_ptr = key.toCharPtr();
  bkt_no = find_bkt(key_ptr);
  delete []key_ptr;

  _lists[bkt_no]->addToBack(record_);
  _occupancy++;

  if(_occupancy >= _fheader._num_bkts * _fheader._num_slots * FILL_FACTOR) {
    GLOBAL_logFileStream << _idString << ": insert; flushing _lists whose total size is " << _occupancy << endl;  
    flush_to_disk();
  }
  return 0;
}

template <class key_type, class data_type, class record_type>
int LinearHash<key_type, data_type, record_type>::update(record_type record_) {
  key_type key;
  char *key_ptr;
  int bkt_no;

  key = record_.get_key();
  key_ptr = key.toCharPtr();
  bkt_no = find_bkt(key_ptr);
  delete []key_ptr;

  int count1 = _lists[bkt_no]->length();
  _lists[bkt_no]->insert_replace(record_);
  int count2 = _lists[bkt_no]->length();
  _occupancy += (count2-count1);

  //GLOBAL_logFileStream << _idString << ": update; inserting " << key << " into _lists[" << bkt_no << "] count1 = " << count1 << ", count2 = " << count2 << endl;
  if(_occupancy >= _fheader._num_bkts * _fheader._num_slots * FILL_FACTOR) {
    GLOBAL_logFileStream << _idString << ": update; flushing _lists whose total size is " << _occupancy << endl;  
    flush_to_disk();
  }
  return 0;
}

template <class key_type, class data_type, class record_type>
int LinearHash<key_type, data_type, record_type>::flush_to_disk() {
  bool split_flag, full_flag, new_flag, insert_flag, inserted;
  BktHeader bkt_hdr;
  record_type record;
  key_type key, key_onfile;
  data_type data, data_onfile;
  char *key_ptr, *key_onfile_ptr, *data_onfile_ptr, *data_ptr, *junk_ptr;
  char *old_bkt_buffer, *new_bkt_buffer, *bkt_buffer, *tmp_bkt_ptr, *old_tmp_ptr, *new_tmp_ptr;
  int temp_bkt_num, bkt_no, data_len, key_len, data_onfile_len, key_onfile_len, list_bkt_num;
  int count, count2, count3, junk_len, bkt_offset, num_bkts, bkt_len;
  struct stat buf;
  struct timeval startTime;
  struct timeval stopTime;
  double insertTime;

  assert(gettimeofday(&startTime, NULL) != -1);
  
  //=================VERIFIED PRINTOUT=============================begin/
  new_flag = false;
  split_flag = false;
  num_bkts = _fheader._num_bkts;

  bkt_len    = _fheader._recd_size * _fheader._num_slots + sizeof(BktHeader);
  bkt_buffer = new char[bkt_len];

  for(temp_bkt_num = 0; temp_bkt_num<num_bkts; temp_bkt_num++) {
    if(_lists[temp_bkt_num]->isEmpty()) continue;

    // reset flags for new bucket
    new_flag = false;
    full_flag = false;
    inserted = false;

    // lseek to new bucket locatioin
    if(goto_bkt(temp_bkt_num) < 0) {
      GLOBAL_logFileStream << _idString << ": flush_to_disk; lseek failed" << endl;
      delete []bkt_buffer;
      return -1;
    }
    
    // remember the bucket offset
    bkt_offset = lseek(_file_desc, 0, SEEK_CUR);
    
    //GLOBAL_logFileStream << _idString 
    //<< ": flush_to_disk; bkt = " << temp_bkt_num 
    //<< ", offset = " << bkt_offset << endl;
    
    // read the bucket
    if(my_read(_file_desc, bkt_buffer, bkt_len) < bkt_len) {
      GLOBAL_logFileStream << _idString << ": flush_to_disk; read of bkt failed" << endl;
      delete []bkt_buffer;
      return -1;
    }

    memcpy((char *)&bkt_hdr, bkt_buffer, sizeof(BktHeader));

    if(bkt_hdr._free_slot == _fheader._num_slots) {
      full_flag = true;
      if(!split_flag) split_flag = true;
    }
    
    _lists[temp_bkt_num]->goFirst();
    while(!full_flag && !_lists[temp_bkt_num]->isOffList()) {
      // get key from record and compute the bkt number
      record = _lists[temp_bkt_num]->getFromCurrent();
      key = record.get_key();
      
      insert_flag = false;
      
      count = 0;
      while(!insert_flag && count < bkt_hdr._free_slot) {
	tmp_bkt_ptr = bkt_buffer + sizeof(BktHeader) + count*_fheader._recd_size;
	memcpy((char *)&key_onfile_len, tmp_bkt_ptr, sizeof(int));
	tmp_bkt_ptr += sizeof(int);

	key_onfile_ptr = new char[key_onfile_len+1];
	memcpy((char *)key_onfile_ptr, tmp_bkt_ptr, key_onfile_len);
	key_onfile_ptr[key_onfile_len] = 0;
	tmp_bkt_ptr += key_onfile_len;

	
	key_onfile = key_type(key_onfile_ptr);
	
	delete []key_onfile_ptr;
	
	if(key_onfile == key) {
	  insert_flag = true;
	  
	  data = record.get_data();
	  data_ptr = data.toCharPtr();
	  data_len = data.get_len();
	  
	  memcpy(tmp_bkt_ptr, (char *)&data_len, sizeof(int));
	  tmp_bkt_ptr += sizeof(int);

	  memcpy(tmp_bkt_ptr, (char *)data_ptr, data_len);
	  tmp_bkt_ptr += data_len;

	  delete []data_ptr;
	}
	count++;
	//if(insert_flag) 
	//GLOBAL_logFileStream << _idString << ": flush_to_disk; " 
	//<< key << " inserted in [" << temp_bkt_num << "].[" << count << "]" << endl;
      }
      
      if(!insert_flag) {
	tmp_bkt_ptr = bkt_buffer + sizeof(BktHeader) + _fheader._recd_size * bkt_hdr._free_slot;
	
	key = record.get_key();
	key_ptr = key.toCharPtr();
	key_len = key.get_len();
	
	data = record.get_data();	  
	data_ptr = data.toCharPtr();
	data_len = data.get_len();
	
	memcpy(tmp_bkt_ptr, (char *)&key_len, sizeof(int));
	tmp_bkt_ptr += sizeof(int);

	memcpy(tmp_bkt_ptr, (char *)key_ptr, key_len);
	tmp_bkt_ptr += key_len;

	memcpy(tmp_bkt_ptr, (char *)&data_len, sizeof(int));
	tmp_bkt_ptr += sizeof(int);

	memcpy(tmp_bkt_ptr, (char *)data_ptr, data_len);
	tmp_bkt_ptr += data_len;

	delete []key_ptr;
	delete []data_ptr;
	
	insert_flag = true;
	if(!new_flag) new_flag = true;
	bkt_hdr._free_slot = bkt_hdr._free_slot + 1;
	
	//GLOBAL_logFileStream << _idString << ": flush_to_disk; " 
	//<< key << " inserted in [" << temp_bkt_num << "].[" << bkt_hdr._free_slot  << "]" << endl;
	
	if(bkt_hdr._free_slot == _fheader._num_slots) {
	  if(!split_flag) split_flag = true;
	  full_flag = true;
	}
      }
      
      if(insert_flag && !inserted) inserted = true;
      _lists[temp_bkt_num]->delCurr();
      _occupancy--;
    }

    // if bucket grew in size
    if(new_flag) {
      memcpy(bkt_buffer, (char *)&bkt_hdr, sizeof(BktHeader));
    
      //GLOBAL_logFileStream << _idString << ": flush_to_disk; bkt grew; writing bkt hdr, offset = " << bkt_offset << endl;
    }
    
    if(inserted) {
      if(lseek(_file_desc, bkt_offset, SEEK_SET) < 0) {
	GLOBAL_logFileStream << _idString << ": flush_to_disk; lseek to bkt offset failed" << endl;
	delete []bkt_buffer;
	return -1;
      }
      
      if(my_write(_file_desc, (char *)bkt_buffer, bkt_len) < bkt_len) { 
	GLOBAL_logFileStream << _idString << ": flush_to_disk; write of bkt failed" << endl;
	delete []bkt_buffer;
	return -1;
      }    
    }
  }
  
  delete []bkt_buffer;

  assert(gettimeofday(&stopTime, NULL) != -1);
  insertTime = 
    (stopTime.tv_sec-startTime.tv_sec) + 
    (stopTime.tv_usec-startTime.tv_usec) * 1e-6;
  GLOBAL_logFileStream << _idString << ": flush_to_disk; flush time = " << insertTime << " seconds" << endl;    


  //=================VERIFIED PRINTOUT=============================end/
  if(split_flag) {
    assert(gettimeofday(&startTime, NULL) != -1);

    // increase number of bkts
    _fheader._num_bkts *= 2;
    //GLOBAL_logFileStream << _idString 
    //<<  ": flush_to_disk; split flag is true and current num bkts = " 
    //<< _fheader._num_bkts << endl;

    _temp = _lists;

    // create new _lists
    num_bkts = _fheader._num_bkts/2;
    _lists = new (List<record_type, key_type> *) [2 * num_bkts];
    for(list_bkt_num=0; list_bkt_num<(2*num_bkts); list_bkt_num++) 
      _lists[list_bkt_num] = new List<record_type, key_type> ();
    _occupancy = 0;

    //GLOBAL_logFileStream << _idString 
    //<<  ": flush_to_disk; created new _lists" << endl;

    // go through each bucket
    bkt_len = _fheader._recd_size * _fheader._num_slots + sizeof(BktHeader);
    old_bkt_buffer = new char[bkt_len];
    new_bkt_buffer = new char[bkt_len];
    memset(new_bkt_buffer, 0, bkt_len);
    for(count = 0; count < num_bkts; count++) {
      // get the bucket information
      if((bkt_offset = lseek(_file_desc, sizeof(HashFileHeader) + count * bkt_len, 
			     SEEK_SET)) < 0) {
	GLOBAL_logFileStream << _idString << ": flush_to_disk; lseek to beginning of bkt failed, offset = " << bkt_offset << endl;
	return -1;
      }
      
      if(my_read(_file_desc, (char *)old_bkt_buffer, bkt_len) < bkt_len) {
	GLOBAL_logFileStream << _idString 
			     << ": flush_to_disk; read of bkt failed, bkt = " << count << endl;
      }
      
      memcpy((char *)&bkt_hdr, old_bkt_buffer, sizeof(BktHeader));


      // rehashing items in bucket
      count3 = 0;
      for(count2 = 0; count2 < bkt_hdr._free_slot; count2++) {
	old_tmp_ptr = old_bkt_buffer + sizeof(BktHeader) + count2 * _fheader._recd_size;

	memcpy((char *)&key_onfile_len, old_tmp_ptr, sizeof(int));
	old_tmp_ptr += sizeof(int);

	
	key_onfile_ptr = new char[key_onfile_len + 1];
	memcpy(key_onfile_ptr, old_tmp_ptr, key_onfile_len);
	key_onfile_ptr[key_onfile_len] = 0;
	old_tmp_ptr += key_onfile_len;

	memcpy((char *)&data_onfile_len, old_tmp_ptr, sizeof(int));
	old_tmp_ptr += sizeof(int);
      
	data_onfile_ptr = new char[data_onfile_len];
	memcpy(data_onfile_ptr, old_tmp_ptr, data_onfile_len);
	old_tmp_ptr += data_onfile_len;

	bkt_no  = find_bkt(key_onfile_ptr);
	record = record_type(key_type(key_onfile_ptr), data_type(data_onfile_ptr));

	if(bkt_no == count) {
	  new_tmp_ptr = new_bkt_buffer + sizeof(BktHeader) + count3 * _fheader._recd_size;

	  count3++;
	  memcpy(new_tmp_ptr, (char *)&key_onfile_len, sizeof(int));
	  new_tmp_ptr += sizeof(int);

	  memcpy(new_tmp_ptr, key_onfile_ptr, key_onfile_len);
	  new_tmp_ptr += key_onfile_len;
	  
	  memcpy(new_tmp_ptr, (char *)&data_onfile_len, sizeof(int));	  
	  new_tmp_ptr += sizeof(int);
	  
	  memcpy(new_tmp_ptr, data_onfile_ptr, data_onfile_len);
	  new_tmp_ptr += data_onfile_len;
	}
	else {
	  _lists[bkt_no]->addToBack(record);
	  _occupancy++;
	}
	delete []key_onfile_ptr;
	delete []data_onfile_ptr;
      } // for(count2..

      // rewrite bucket
      bkt_hdr._free_slot = count3;
      bkt_hdr._over_flow = -1;
      memcpy(new_bkt_buffer, (char *)&bkt_hdr, sizeof(BktHeader));

      if(lseek(_file_desc, bkt_offset, SEEK_SET) < 0) {
	GLOBAL_logFileStream << _idString << ": flush_to_disk; lseek failed, offset = " << bkt_offset << endl;
	delete []old_bkt_buffer;
	delete []new_bkt_buffer;		
	return -1;
      }

      if(my_write(_file_desc, new_bkt_buffer, bkt_len) < bkt_len) {
	GLOBAL_logFileStream << _idString << ": flush_to_disk; unable to write bkt" << endl;
	delete []old_bkt_buffer;
	delete []new_bkt_buffer;       
	return -1;
      }

      //GLOBAL_logFileStream << _idString << ": flush_to_disk; rewrote bkt " << count << ", new occupancy " << count3 << endl;
    }
    
    delete []old_bkt_buffer;
    delete []new_bkt_buffer;

    for(temp_bkt_num=0; temp_bkt_num<num_bkts; temp_bkt_num++) {
      _temp[temp_bkt_num]->goFirst();
      while(!_temp[temp_bkt_num]->isOffList()) {
	record = _temp[temp_bkt_num]->getFromCurrent();
	key = record.get_key();
	key_ptr = key.toCharPtr();
	bkt_no = find_bkt(key_ptr);
	_lists[bkt_no]->addToBack(record);
	_occupancy++;
	_temp[temp_bkt_num]->delCurr();
	delete []key_ptr;
      }
      assert(_temp[temp_bkt_num]->isEmpty());
      delete _temp[temp_bkt_num];
    }
    delete []_temp;
    _temp = NULL;

    //GLOBAL_logFileStream << _idString << ": flush_to_disk; moved list items from _temp to _lists" << endl;

    assert(count == _fheader._num_bkts/2);
    
    // go to end of file
    if((bkt_offset = lseek(_file_desc, sizeof(HashFileHeader) + count * (sizeof(BktHeader) + _fheader._num_slots * _fheader._recd_size), 
			   SEEK_SET)) < 0) {
      GLOBAL_logFileStream << _idString << ": flush_to_disk; lseek to end of all bkts failed" << endl;
      return -1;    
    }

    if(stat(_file_name, &buf) == 0 && bkt_offset != buf.st_size) {
      GLOBAL_logFileStream << _idString << ": flush_to_disk; offset = " << bkt_offset << ", file size = " <<  buf.st_size << endl;
      assert(false);
    }

    // add new bkts
    bkt_hdr._over_flow = -1;
    bkt_hdr._free_slot = 0;

    junk_len = _fheader._num_slots * _fheader._recd_size + sizeof(BktHeader);
    junk_ptr = new char[junk_len];
    memcpy(junk_ptr, (char *)&bkt_hdr, sizeof(BktHeader));
    memset(junk_ptr + sizeof(BktHeader), '0', junk_len - sizeof(BktHeader));

    for(count = 0; count < num_bkts; count++) {
      if(my_write(_file_desc, junk_ptr, junk_len) < junk_len) {
	GLOBAL_logFileStream << _idString << ": flush_to_disk; bkt slots write failed" << endl;
	return -1;
      }
    }
    //GLOBAL_logFileStream << _idString << ": flush_to_disk; added buckets " << num_bkts << "-" << num_bkts+count-1 << endl;
    delete []junk_ptr;

    assert(gettimeofday(&stopTime, NULL) != -1);
    insertTime = 
      (stopTime.tv_sec-startTime.tv_sec) + 
      (stopTime.tv_usec-startTime.tv_usec) * 1e-6;
    GLOBAL_logFileStream << _idString << ": flush_to_disk; split time = " << insertTime << " seconds" << endl;    
    

    return 1; // this indicates to the caller that _lists[bkt_no] is not empty for some value of bkt_no
  }
  else {
    assert(_occupancy == 0);
    return 0;
  }
}

