#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <iostream.h>
#include "../index/InvertedFile5.h" //NUM_BUCKETS, NUM_SLOTS
#include "../index/Index5.h" 
#include "../index/LinearHash3.h"
#include "../index/key_offset_record.h"
#include <stdlib.h>
#include <math.h>

#define NUM_KEYWORD_BUCKETS (15013) //151
#define NUM_KEYWORD_SLOTS   (100)

ofstream GLOBAL_logFileStream;

main(int argc, char* argv[]) {
  struct timeval startTime;
  struct timeval stopTime;
  double createTime;
  if(argc < 3) {
    cout << "Usage: " << argv[0] << " <in_file> <rec_size> " << endl;
    exit(0);
  }
  assert(gettimeofday(&startTime, NULL) != -1);
  GLOBAL_logFileStream.open("create_index.log");
  char *in_file = argv[1], *key;
  int in_fd = open(in_file, O_RDONLY), eof_offset, offset, keylen, listlen, rec_size = atoi(argv[2]), count = 0, index_num = 0;
  double percent;
  RecHeader rh;
  key_offset_record kor;
  char *out_file = new char[strlen(in_file) + strlen(".idx.") + 10 + 1];
  sprintf(out_file, "%s.idx.%d", in_file, index_num);  
  char *id_string = new char[strlen("Final.LH.") + 10 + 1];
  sprintf(id_string, "Final.LH.%d", index_num);
  index_num++;
  LinearHash<String, Int, key_offset_record> *index;
  index = new LinearHash<String, Int, key_offset_record>(out_file, NUM_KEYWORD_BUCKETS, NUM_KEYWORD_SLOTS, sizeof(int) + String::get_max_len(), id_string);

  if(rec_size > 0) rec_size =  BASIC_REC_SIZE * (int) pow(3, rec_size - 1);

  eof_offset = lseek(in_fd, 0, SEEK_END);
  offset = sizeof(FileHeader);
  if(lseek(in_fd, offset, SEEK_SET) < 0) return 0;

  while(offset < eof_offset) {
    if(read(in_fd, (char *)&rh, sizeof(RecHeader)) < sizeof(RecHeader)) return 0;
    if(read(in_fd, (char *)&keylen, sizeof(int)) < sizeof(int)) return 0;

    key = new char[keylen + 1];    
    if(read(in_fd, (char *)key, keylen) < keylen) return 0;
    key[keylen] = 0;

    if(read(in_fd, (char *)&listlen, sizeof(int)) < sizeof(int)) return 0;
    
    kor = key_offset_record(String(key), Int(offset));
    if(keylen <= String::get_max_len()) {
      if((index_num == 1 && strcmp(key, "g") >= 0) || 
	 (index_num == 2 && strcmp(key, "n") >= 0) || 
	 (index_num == 3 && strcmp(key, "s") >= 0)) {
	delete index;
	assert(gettimeofday(&stopTime, NULL) != -1);
	createTime = 
	  (stopTime.tv_sec-startTime.tv_sec) + 
	  (stopTime.tv_usec-startTime.tv_usec) * 1e-6;
	GLOBAL_logFileStream << id_string << ": total index creation time " << createTime << " seconds" << endl;    

	assert(gettimeofday(&startTime, NULL) != -1);

	sprintf(out_file, "%s.idx.%d", in_file, index_num);  
	sprintf(id_string, "Final.LH.%d", index_num);
	index_num++;
	index = new LinearHash<String, Int, key_offset_record>(out_file, NUM_KEYWORD_BUCKETS, NUM_KEYWORD_SLOTS, sizeof(int) + String::get_max_len(), id_string);
	count = 0;
      }	
      double cpu_time, io_time;
      if(index->insert(kor, cpu_time, io_time) == -1) exit(0);
    }
    offset += (sizeof(RecHeader) + ((rec_size > 0) ? rec_size : (2 * sizeof(int) + keylen + listlen)));
    percent = (float)offset/float(eof_offset) * 100;
    cout << "Inserting key = " << key << "[" << ++count /*percent << "%" */ << "]" << endl;
    delete []key;
    
    if(lseek(in_fd, offset, SEEK_SET) < 0) return 0;
  }

  close(in_fd);
  delete index;
  assert(gettimeofday(&stopTime, NULL) != -1);
  createTime = 
    (stopTime.tv_sec-startTime.tv_sec) + 
    (stopTime.tv_usec-startTime.tv_usec) * 1e-6;
  GLOBAL_logFileStream << id_string  << ": total index creation time " << createTime << " seconds" << endl;    
  
  delete []out_file;
  delete []id_string;
  GLOBAL_logFileStream.close();
  return 1;
}

